# -*- coding: utf-8 -*-
from .analyzer import *
from .utils import *
from .event_loop import *
from .executor import *
from .formdata import *
from .futures import *
from .http import *
from .ios import *
from .websocket import *
from .export import *

__all__ = (
    *analyzer.__all__,
    *utils.__all__,
    *event_loop.__all__,
    *executor.__all__,
    *formdata.__all__,
    *futures.__all__,
    *http.__all__,
    *ios.__all__,
    *websocket.__all__,
    *export.__all__,
)



# -*- coding: utf-8 -*-
__all__ = ('ChooseMenu', )

from ...backend.futures import CancelledError
from ...backend.utils import copy_docs
from ...discord.emoji import BUILTIN_EMOJIS
from ...discord.parsers import InteractionEvent
from ...discord.message import Message
from ...discord.channel import ChannelTextBase
from ...discord.exceptions import DiscordException, ERROR_CODES
from ...discord.embed import Embed

from .bases import GUI_STATE_READY, GUI_STATE_SWITCHING_PAGE, GUI_STATE_CANCELLING, GUI_STATE_SWITCHING_CTX, \
    GUI_STATE_VALUE_TO_NAME, PaginationBase
from .utils import Timeouter


class ChooseMenu(PaginationBase):
    """
    Familiar to ``Pagination``, but instead of just displaying multiple pages of text, it allows the user to select
    a displayed option.
    
    The class allows modifications and closing it's representations for every user. Also works at private channels.
    
    Picks up on reaction additions and on reaction deletions as well and removes the added reactions on if has
    permission, which might be missing, like in DM-s.
    
    Attributes
    ----------
    _canceller : `None` or `function`
        The function called when the ``ChooseMenu`` is cancelled or when it expires. This is a onetime use and after
        it was used, is set as `None`.
    
    _task_flag : `int`
        A flag to store the state of the ``ChooseMenu``.
        
        Possible values:
        +---------------------------+-------+-----------------------------------------------------------------------+
        | Respective name           | Value | Description                                                           |
        +===========================+=======+=======================================================================+
        | GUI_STATE_READY           | 0     | The ChooseMenu does nothing, is ready to be used.                     |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_SWITCHING_PAGE  | 1     | The ChooseMenu is currently changing it's page.                       |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_CANCELLING      | 2     | The ChooseMenu is currently changing it's page, but it was cancelled  |
        |                           |       | meanwhile.                                                            |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_CANCELLED       | 3     | The ChooseMenu is, or is being cancelled right now.                   |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_SWITCHING_CTX   | 4     | The ChooseMenu is switching context. Not used by the default class,   |
        |                           |       | but expected.                                                         |
        +---------------------------+-------+-----------------------------------------------------------------------+
    
    _timeouter : `None` or ``Timeouter``
        Executes the timing out feature on the ``ChooseMenu``.
    
    channel : ``ChannelTextBase`` instance
        The channel where the ``ChooseMenu`` is executed.
    
    client : ``Client``
        The client who executes the ``ChooseMenu``.
    message : `None` or ``Message``
        The message on what the ``ChooseMenu`` is executed.
    
    check : `None` or `callable`
        A callable what decides whether the ``ChooseMenu`` should process a received reaction event. Defaults to
        `None`.
        
        Should accept the following parameters:
        +-----------+---------------------------------------------------+
        | Name      | Type                                              |
        +===========+===================================================+
        | event     | ``ReactionAddEvent`` or ``ReactionDeleteEvent``   |
        +-----------+---------------------------------------------------+
        
        Note, that ``ReactionDeleteEvent`` is only given, when the client has no `manage_messages` permission.
        
        Should return the following values:
        +-------------------+-----------+
        | Name              | Type      |
        +===================+===========+
        | should_process    | `bool`    |
        +-------------------+-----------+
    
    embed : ``EmbedBase`` instance
        An embed base, what's description and footer will be rendered with the given choices and with information
        about the respective page.

    selected : `int`
        The currently selected option of the ``ChooseMenu``.
    
    choices : `indexable` of `Any`
        An indexable container, what stores the displayable choices.
        
        It's elements's type can be different from each other, and different structures act differently as well.
        There are the following cases:
        
        - If an element is `str` instance, then it will be used as an option's title and when selecting it, only that
            variable will be passed to the respective function when selected.
        
        - If an element is neither `str` or `tuple`, then it's `repr` will be used as an option's title, and only that
            variable will be passed to the respective function when selected.
        
        - If an element is `tuple` instance, then it's first element will be displayed as title. If it is `str`, then
            will be just simply added, however if not, then it's `repr` will be used. If selecting a `tuple` option,
            then it's element will be passed to the respective function.
    
    timeout : `float`
        The timeout of the ``ChooseMenu`` in seconds.
    
    prefix : `None` or `str`
        A prefix displayed before each option.
    
    selector : `async-callable`
        An `async-callable`, what is ensured when an option is selected.
        
        If the ``ChooseMenu`` is created only with `1` option, then it is ensured initially instead of creating the
        ``ChooseMenu`` itself. At this case, if `message` was not given (or given as `None`), then the `message`
        passed to the `selector` will be `None` as well.
        
        At least 3 parameters are passed to the `selector`:
        +-------------------+-------------------------------+
        | Respective name   | Type                          |
        +===================+===============================+
        | client            | ``Client``                    |
        +-------------------+-------------------------------+
        | channel           | ``ChannelTextBase`` instance  |
        +-------------------+-------------------------------+
        | message           | ``Message`` or `None`         |
        +-------------------+-------------------------------+
        
        The rest of the parameters depend on the respective choice (an elements of ``choices``). If the element is a
        `tuple` instance, then it's element will be passed, however if the choice is any other type, then only that
        object will be passed.
    
    Class Attributes
    ----------------
    UP : ``Emoji`` = `BUILTIN_EMOJIS['arrow_up_small']`
        The emoji used to move on the displayed option one above.
    DOWN : ``Emoji`` = `BUILTIN_EMOJIS['arrow_down_small']`
        The emoji used to move on the displayed option one under.
    LEFT : ``Emoji`` = `BUILTIN_EMOJIS['arrow_backward']`
        The emoji used to move on the previous page.
    RIGHT : ``Emoji`` = `BUILTIN_EMOJIS['arrow_forward']`
        The emoji used to move on the next page.
    SELECT : ``Emoji`` = `BUILTIN_EMOJIS['ok']`
        The emoji used to select an option.
    CANCEL : ``Emoji`` = `BUILTIN_EMOJIS['x']`
        The emoji used to cancel the ``ChooseMenu``.
    EMOJIS_RESTRICTED : `tuple` (`Emoji`, `Emoji`, `Emoji`, `Emoji`) = `(UP, DOWN, SELECT, CANCEL)`
        Restricted emojis, added when the choose menu has only options for 1 page.
    EMOJIS : `tuple` (`Emoji`, `Emoji`, `Emoji`, `Emoji`, `Emoji`, `Emoji`) = `(UP, DOWN, LEFT, RIGHT, SELECT, CANCEL)`
        Emojis added to the choose menu.
    """
    UP = BUILTIN_EMOJIS['arrow_up_small']
    DOWN = BUILTIN_EMOJIS['arrow_down_small']
    LEFT = BUILTIN_EMOJIS['arrow_backward']
    RIGHT = BUILTIN_EMOJIS['arrow_forward']
    SELECT = BUILTIN_EMOJIS['ok']
    CANCEL = BUILTIN_EMOJIS['x']
    EMOJIS_RESTRICTED = (UP, DOWN, SELECT, CANCEL)
    EMOJIS = (UP, DOWN, LEFT, RIGHT, SELECT, CANCEL)
    
    __slots__ = ('check', 'embed', 'selected', 'choices', 'timeout', 'prefix', 'selector')
    
    async def __new__(cls, client, channel, choices, selector, embed=Embed(), *, timeout=240., message=None,
            prefix=None, check=None):
        """
        Creates a new choose menu with the given parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who executes the ``ChooseMenu``.
        channel : ``ChannelTextBase`` instance or ``Message``
            The channel where the ``ChooseMenu`` is executed. Pass it as a ``Message`` instance to send a reply.
            
            If given as ``InteractionEvent``, then will acknowledge it and create a new message with it as well.
            Although will not acknowledge it if `message` is given.
        
        choices : `indexable` of `Any`
            An indexable container, what stores the displayable choices.
            
            It's elements's type can be different from each other, and different structures act differently as well.
            There are the following cases:
            
            - If an element is `str` instance, then it will be used as an option's title and when selecting it, only
                that variable will be passed to the respective function when selected.
            
            - If an element is neither `str` or `tuple`, then it's `repr` will be used as an option's title, and only
                that variable will be passed to the respective function when selected.
            
            - If an element is `tuple` instance, then it's first element will be displayed as title. If it is `str`,
                then will be just simply added, however if not, then it's `repr` will be used. If selecting a `tuple`
                option, then it's element will be passed to the respective function.
        selector : `async-callable`
            An `async-callable`, what is ensured when an option is selected.
            
            If the ``ChooseMenu`` is created only with `1` option, then it is ensured initially instead of creating
            the ``ChooseMenu`` itself. At this case, if `message` was not given (or given as `None`), then the
            `message` passed to the `selector` will be `None` as well.
            
            At least 3 parameters are passed to the `selector`:
            +-------------------+-----------------------------------------------------------+
            | Respective name   | Type                                                      |
            +===================+===========================================================+
            | client            | ``Client``                                                |
            +-------------------+-----------------------------------------------------------+
            | channel           | ``ChannelTextBase``, ``Message`` or ``InteractionEvent``  |
            +-------------------+-----------------------------------------------------------+
            | message           | ``Message`` or `None`                                     |
            +-------------------+-----------------------------------------------------------+
            
            The rest of the parameters depend on the respective choice (an elements of ``choices``). If the element is a
            `tuple` instance, then it's element will be passed, however if the choice is any other type, then only that
            object will be passed.
        embed : ``EmbedBase``
            An embed base, what's description and footer will be rendered with the given choices and with information
            about the respective page. Defaults to an empty ``Embed`` instance.
        timeout : `float`, Optional (Keyword only)
            The timeout of the ``ChooseMenu`` in seconds. Defaults to `240.0`.
        message : `None` or ``Message``, Optional (Keyword only)
            The message on what the ``ChooseMenu`` will be executed. If not given a new message will be created.
            Defaults to `None`.
        prefix : `None` or `str`, Optional (Keyword only)
            A prefix displayed before each option. Defaults to `None`.
        check : `None` or `callable`, Optional (Keyword only)
            A callable what decides whether the ``ChooseMenu`` should process a received reaction event. Defaults to
            `None`.
            
            Should accept the following parameters:
            +-----------+---------------------------------------------------+
            | Name      | Type                                              |
            +===========+===================================================+
            | event     | ``ReactionAddEvent`` or ``ReactionDeleteEvent``   |
            +-----------+---------------------------------------------------+
            
            Note, that ``ReactionDeleteEvent`` is only given, when the client has no `manage_messages` permission.
            
            Should return the following values:
            +-------------------+-----------+
            | Name              | Type      |
            +===================+===========+
            | should_process    | `bool`    |
            +-------------------+-----------+
        
        Returns
        -------
        self : `None` or ``ChooseMenu``
            If `choices`'s length is less than `2`, then returns `None`.
        
        Raises
        ------
        TypeError
            `channel`'s type is incorrect.
        ValueError
            If `prefix` was not given as `None` and it's length is over `64` characters.
        """
        if (prefix is not None) and (len(prefix) > 100):
            raise ValueError(f'Please pass a prefix, what is shorter than 100 characters, got {prefix!r}.')
        
        if isinstance(channel, ChannelTextBase):
            target_channel = channel
            received_interaction = False
        elif isinstance(channel, Message):
            target_channel = channel.channel
            received_interaction = False
        elif isinstance(channel, InteractionEvent):
            target_channel = channel.channel
            received_interaction = True
        else:
            raise TypeError(f'`channel` can be given only as `{ChannelTextBase.__name__}`, `{Message.__name__}` '
                f'of as {InteractionEvent.__name__} instance, got {channel.__class__.__name__}.')
        
        result_ln = len(choices)
        if result_ln < 2:
            if result_ln == 1:
                choice = choices[0]
                if isinstance(choice, tuple):
                    coro = selector(client, channel, message, *choice)
                else:
                    coro = selector(client, channel, message, choice)
                await coro
            return None
        
        self = object.__new__(cls)
        self.check = check
        self.client = client
        self.channel = target_channel
        self.choices = choices
        self.selector = selector
        self.selected = 0
        self._canceller = cls._canceller_function
        self._task_flag = GUI_STATE_READY
        self.message = message
        self.timeout = timeout
        self._timeouter = None
        self.prefix = prefix
        self.embed = embed
        
        try:
            if message is None:
                if received_interaction:
                    if not channel.is_acknowledged():
                        await client.interaction_response_message_create(channel)
                    
                    message = await client.interaction_followup_message_create(channel, embed=self._render_embed())
                else:
                    message = await client.message_create(channel, embed=self._render_embed())
                self.message = message
            else:
                await client.message_edit(message, embed=self._render_embed())
        except BaseException as err:
            self.cancel(err)
            if isinstance(err, ConnectionError):
                return self
            
            if isinstance(err, DiscordException):
                if err.code in (
                        ERROR_CODES.unknown_message, # message deleted
                        ERROR_CODES.unknown_channel, # message's channel deleted
                        ERROR_CODES.max_reactions, # reached reaction 20, some1 is trolling us.
                        ERROR_CODES.invalid_access, # client removed
                        ERROR_CODES.invalid_permissions, # permissions changed meanwhile
                        ERROR_CODES.cannot_message_user, # user has dm-s disallowed
                            ):
                    return self
            
            raise
        
            
        if not target_channel.cached_permissions_for(client).can_add_reactions:
            self.cancel(PermissionError())
            return self
        
        
        try:
            for emoji in (self.EMOJIS if (len(choices) > 10) else self.EMOJIS_RESTRICTED):
                await client.reaction_add(message, emoji)
        except BaseException as err:
            self.cancel(err)
            if isinstance(err, ConnectionError):
                return self
            
            if isinstance(err, DiscordException):
                if err.code in (
                        ERROR_CODES.unknown_message, # message deleted
                        ERROR_CODES.unknown_channel, # message's channel deleted
                        ERROR_CODES.max_reactions, # reached reaction 20, some1 is trolling us.
                        ERROR_CODES.invalid_access, # client removed
                        ERROR_CODES.invalid_permissions, # permissions changed meanwhile
                            ):
                    return self
            
            raise
        
        self._timeouter = Timeouter(self, timeout)
        client.events.reaction_add.append(message, self)
        client.events.reaction_delete.append(message, self)
        return self
    
    
    def _render_embed(self):
        """
        Renders the choose menu's embed's description with it's choices of the respective page and it's footer
        with page information.
        
        Returns
        -------
        embed : ``Embed`` (or any compatible)
            The rendered embed.
        """
        selected = self.selected
        choices = self.choices
        index = (selected//10)*10
        end = index+10
        if len(choices) < end:
            end = len(choices)
        
        parts = []
        prefix = self.prefix
        left_length = 195
        if (prefix is not None):
            left_length -= len(prefix)
        
        while True:
            title = choices[index]
            if isinstance(title,tuple):
                if not title:
                    title = ''
                else:
                    title = title[0]
            
            if not isinstance(title,str):
                title = str(title)
            
            if len(title) > left_length:
                space_position = title.rfind(' ', left_length-25, left_length)
                if space_position == -1:
                    space_position = left_length-3
                
                title = title[:space_position]+'...'
            
            if index == selected:
                if (prefix is not None):
                    parts.append('**')
                    parts.append(prefix)
                    parts.append('** ')
                parts.append('**')
                parts.append(title)
                parts.append('**\n')
            else:
                if (prefix is not None):
                    parts.append(prefix)
                    parts.append(' ')
                parts.append(title)
                parts.append('\n')
            
            index +=1
            if index == end:
                break
        
        embed = self.embed
        embed.description = ''.join(parts)
        
        current_page = (selected//10)+1
        limit = len(choices)
        page_limit = (limit//10)+1
        start = end-9
        if start < 1:
            start = 1
        if end == len(choices):
            end -= 1
        limit -= 1
        
        embed.add_footer(f'Page {current_page}/{page_limit}, {start} - {end} / {limit}, selected: {selected+1}')
        return embed
    
    @copy_docs(PaginationBase.__call__)
    async def __call__(self, client, event):
        if event.user.is_bot:
            return
        
        if (event.emoji not in (self.EMOJIS if len(self.choices)>10 else self.EMOJIS_RESTRICTED)):
            return
        
        if (event.delete_reaction_with(client) == event.DELETE_REACTION_NOT_ADDED):
            return
        
        check = self.check
        if (check is not None):
            try:
                should_continue = check(event)
            except BaseException as err:
                await client.events.error(client, f'{self!r}.__call__', err)
                return
            
            if not should_continue:
                return
        
        task_flag = self._task_flag
        if task_flag != GUI_STATE_READY:
            if task_flag == GUI_STATE_SWITCHING_PAGE:
                if event.emoji is self.CANCEL:
                    self._task_flag = GUI_STATE_CANCELLING
                return
            
            # ignore GUI_STATE_CANCELLED and GUI_STATE_SWITCHING_CTX
            return
        
        message = self.message
        
        while True:
            emoji = event.emoji
            if emoji is self.UP:
                selected = self.selected-1
                break
            
            if emoji is self.DOWN:
                selected = self.selected+1
                break
            
            if emoji is self.LEFT:
                selected = self.selected-10
                break
            
            if emoji is self.RIGHT:
                selected = self.selected+10
                break
            
            if emoji is self.CANCEL:
                self.cancel(CancelledError())
                return
            
            if emoji is self.SELECT:
                self._task_flag = GUI_STATE_SWITCHING_CTX
                self.cancel()
                
                try:
                    if self.channel.cached_permissions_for(client).can_manage_messages:
                        await client.reaction_clear(message)
                    
                    else:
                        for emoji in self.EMOJIS:
                            await client.reaction_delete_own(message, emoji)
                except BaseException as err:
                    self.cancel(err)
                    if isinstance(err, ConnectionError):
                        # no internet
                        return
                    
                    if isinstance(err, DiscordException):
                        if err.code in (
                                ERROR_CODES.unknown_message, # message already deleted
                                ERROR_CODES.unknown_channel, # channel deleted
                                ERROR_CODES.invalid_access, # client removed
                                ERROR_CODES.invalid_permissions, # permissions changed meanwhile
                                    ):
                            return
                    
                    await client.events.error(client, f'{self!r}.__call__', err)
                    return
                
                selector = self.selector
                try:
                    choice = self.choices[self.selected]
                    channel = self.channel
                    if isinstance(choice, tuple):
                        coro = selector(client, channel, message, *choice)
                    else:
                        coro = selector(client, channel, message, choice)
                    await coro
                except BaseException as err:
                    await client.events.error(client, f'{self!r}.__call__ when calling {selector!r}', err)
                return
            
            return
        
        if selected < 0:
            selected = 0
        elif selected >= len(self.choices):
            selected = len(self.choices)-1
        
        if self.selected == selected:
            return
        
        self.selected = selected
        self._task_flag = GUI_STATE_SWITCHING_PAGE
        try:
            await client.message_edit(message, embed=self._render_embed())
        except BaseException as err:
            self.cancel(err)
            
            if isinstance(err, ConnectionError):
                # no internet
                return
            
            if isinstance(err, DiscordException):
                if err.code in (
                        ERROR_CODES.unknown_message, # message already deleted
                        ERROR_CODES.unknown_channel, # message's channel deleted
                        ERROR_CODES.invalid_access, # client removed
                            ):
                    return
            
            # We definitely do not want to silence `ERROR_CODES.invalid_form_body`
            await client.events.error(client, f'{self!r}.__call__', err)
            return
        
        if self._task_flag == GUI_STATE_CANCELLING:
            self.cancel(CancelledError())
            return
        
        self._task_flag = GUI_STATE_READY
        
        timeouter = self._timeouter
        if (timeouter is not None):
            timeouter.set_timeout(self.timeout)


    @copy_docs(PaginationBase.__repr__)
    def __repr__(self):
        repr_parts = [
            '<', self.__class__.__name__,
            ' client=', repr(self.client),
            ', channel=', repr(self.channel),
            ', state='
        ]
        
        task_flag = self._task_flag
        repr_parts.append(repr(task_flag))
        repr_parts.append(' (')
        
        task_flag_name = GUI_STATE_VALUE_TO_NAME[task_flag]
        
        repr_parts.append(task_flag_name)
        repr_parts.append(')')
        
        # Third party things go here
        repr_parts.append(', choices=')
        repr_parts.append(repr(self.choices))
        repr_parts.append(', selected=')
        repr_parts.append(repr(self.selected))
        repr_parts.append(', selector=')
        repr_parts.append(repr(self.selector))
        
        prefix = self.prefix
        if (prefix is not None):
            repr_parts.append(', prefix=')
            repr_parts.append(repr(prefix))
        
        repr_parts.append('>')
        return ''.join(repr_parts)



# -*- coding: utf-8 -*-
__all__ = ('DocString', )
import re, sys

from .graver import build_graves_on_subsection, GravedListing, GravedListingElement, GravedTable, GravedDescription, \
    GravedCodeBlock, DocWarning, GravedAttributeDescription, GravedBlockQuote

SECTION_NAME_RP = re.compile('(?:[A-Z][a-z]*)(?: [A-Z][a-z]*)*')
SECTION_UNDERLINE_RP = re.compile('[\-]+')
TABLE_BORDER_PATTERN = '\+(?:[\-=]+\+)+'
TABLE_TEXT_PATTERN = '\|(?:[^\|]+\|)+'
TABLE_BORDER_RP = re.compile(TABLE_BORDER_PATTERN)
TABLE_TEXT_RP = re.compile(TABLE_TEXT_PATTERN)
TABLE_ANY_RP = re.compile(f'{TABLE_BORDER_PATTERN}|{TABLE_TEXT_PATTERN}')
TABLE_TEXT_SPLITTER = re.compile('(?:\| )?(.*?) \|')
LISTING_HEAD_LINE_RP = re.compile('[\-]+[ \t]*(.*)')

ATTRIBUTE_SECTION_NAME_RP = re.compile('.*?attributes?', re.I)
ATTRIBUTE_NAME_RP = re.compile('([A-Za-z_][0-9A-Za-z_]*) *([\:\(]) *')

del TABLE_BORDER_PATTERN
del TABLE_TEXT_PATTERN

RIGHT_STRIP_IGNORE = '\\ \t\n'

def remove_indents(lines):
    """
    Removes tailing characters from the right, the base indention and the empty lines from the start and from the end.
    
    Parameters
    ----------
    lines : `list` of `str`
        Input lines.
    
    Returns
    -------
    lines : `None` or (`list` of `str`)
    """
    if not lines:
        return None
    
    # remove tailing characters from the right side
    for index in range(len(lines)):
        lines[index] = lines[index].rstrip(RIGHT_STRIP_IGNORE)
    
    # First line might not be indented, so check it meanwhile it is first.
    first_line = lines[0]
    if first_line and (first_line[0] not in (' ', '\t')):
        first_line_filled = True
    else:
        first_line_filled = False
    
    # Remove empty lines from the end
    while True:
        if not lines:
            return None
        
        if lines[-1]:
            break
        
        del lines[-1]
        continue
    
    # Remove empty lines from the start
    while True:
        if lines[0]:
            break
        
        del lines[0]
        continue
    
    # Collect line indexes, which have content
    content_line_indexes = []
    for index in range(len(lines)):
        line = lines[index]
        if line:
            content_line_indexes.append(index)
    
    # If first line has any content, ignore it when removing indents.
    if first_line_filled and content_line_indexes[0] == 0:
        del content_line_indexes[0]
    
    # Do not remove indents if there is non.
    if content_line_indexes:
        # Calculate indent size
        ignore_index = 0
        
        while True:
            next_char = lines[content_line_indexes[0]][ignore_index]
            if next_char not in ('\t', ' '):
                break
            
            for index in content_line_indexes[1:]:
                char = lines[index][ignore_index]
                if char != next_char:
                    break
                
                continue
            else:
                ignore_index +=1
                continue
            
            break
        
        # Remove indents
        if ignore_index:
            for index in range(first_line_filled, len(lines)):
                line = lines[index]
                
                lines[index] = line[ignore_index:]
                continue
    
    return lines
    
def parse_sections(lines):
    """
    Parses out sections from the given lines.
    
    Parameters
    ----------
    lines : `list` of `str`
        Unindented lines of a docstring.
    
    Returns
    -------
    sections : `list` of `tuple` (`str`, `str`)
    """
    sections = []
    
    section_name = None
    section_lines = []
    
    index = 0
    limit = len(lines)
    
    while True:
        line = lines[index]
        
        # If there are at least 2 lines and we can match:
        if (limit-index) > 2 and (SECTION_NAME_RP.fullmatch(line) is not None) and \
                (SECTION_UNDERLINE_RP.fullmatch(lines[index+1]) is not None):
            # Get the current line out.
            while section_lines:
                if section_lines[-1]:
                    sections.append((section_name, section_lines))
                    section_lines = []
                    break
                
                del section_lines[-1]
                continue
            
            section_name = line
            index += 1
        else:
            section_lines.append(line)
        
        index+=1
        
        if index >= limit:
            while section_lines:
                if section_lines[-1]:
                    sections.append((section_name, section_lines))
                    break
                
                del section_lines[-1]
                continue
            
            break
    
    return sections

def detect_table(lines, index, limit):
    """
    Detects the lines of a table.
    
    Parameters
    ----------
    lines : `list` of `str`
        The lines of the section.
    index : `int`
        The starting index of the table.
    limit : `int`
        The last element's index.
    
    Returns
    -------
    index : `int`
        The index where the table is over. If there was no table, returns the initial index.
    """
    line = lines[index]
    if index+2 <= limit:
        if (TABLE_ANY_RP.fullmatch(line) is not None):
            line2 = lines[index+1]
            if (TABLE_ANY_RP.fullmatch(line2) is not None):
                index += 2
                while True:
                    if index == limit:
                        break
                    
                    line = lines[index]
                    if (TABLE_ANY_RP.fullmatch(line) is not None):
                        index +=1
                        continue
                    
                    break
    
    return index

class TextTable:
    """
    Represents a table inside of a docstring.
    
    Attributes
    ----------
    _array : `list` of (`str`, `None`)
        The elements of the table.
    _size : `tuple` (`int`, `int`)
        The dimensions of the table.
    """
    __slots__ = ('_array', '_size', )
    def __new__(cls, lines, start, end):
        """
        Creates a new table from the given lines's start:end range.
        
        Parameters
        ----------
        lines : `list` of `str`
            Line to create the table from.
        start : `int`
            The first line's index, where the table starts.
        end : `int`
            The 1 after the last line's index of the table.
        
        Returns
        -------
        self : `None` or ``TextTable``
            Returns `None` if would create an empty table.
        """
        splitted_lines = []
        # Split the lines. If a line is a border line, then add `None` to it, else a list of parts.
        while True:
            if start == end:
                break
            
            line = lines[start]
            start +=1
            
            if (TABLE_BORDER_RP.fullmatch(line) is not None):
                splitted_lines.append(None)
                continue
            
            splitted_line = TABLE_TEXT_SPLITTER.findall(line)
            # GOTO
            while True:
                if not splitted_line:
                    # Leave immediately if empty
                    break
                
                if not splitted_line[-1]:
                    # Remove last element, if empty.
                    del splitted_line[-1]
                
                if not splitted_line:
                    break
                
                if not splitted_line[0]:
                    # Remove first element if empty
                    del splitted_line[0]
                    
                if not splitted_line:
                    break
                
                # Replace empty elements with None and also strip the lines down as well.
                for index in range(len(splitted_line)):
                    splitted_part = splitted_line[index]
                    splitted_part = splitted_part.lstrip().rstrip(RIGHT_STRIP_IGNORE)
                    if not splitted_part:
                        splitted_part = None
                    
                    splitted_line[index] = splitted_part
                
                break
            
            splitted_lines.append(splitted_line)
        
        # Lets calculate how long is the longest line.
        longest = 0
        for splitted_part in splitted_lines:
            if splitted_part is None:
                continue
            
            actual_length = len(splitted_part)
            if actual_length > longest:
                longest = actual_length
                continue
        
        # If the longest is 0, we can leave
        if (not longest):
            return None
        
        # Fill up every line what is shorter than the longest with None
        for splitted_line in splitted_lines:
            if splitted_line is None:
                continue
            
            for _ in range(len(splitted_line), longest):
                splitted_line.append(None)
        
        processed_lines = []
        
        index = 0
        limit = len(splitted_lines)
        
        processed_line = [None for _ in range(longest)]
        
        while True:
            if index == limit:
                only_none = True
                for part_index in range(longest):
                    processed_part = processed_line[part_index]
                    if processed_part is None:
                        continue
                    
                    only_none = False
                    if type(processed_part) is list:
                        processed_part = ' '.join(processed_part)
                    
                    processed_line[part_index] = processed_part
                
                if only_none:
                    break
                
                processed_lines.append(processed_line)
                break
            
            splitted_line = splitted_lines[index]
            index +=1
            
            if splitted_line is None:
                # If we are at a border line, join the continuous lines together.
                # If every continuous line is empty, do not add them.
                only_none = True
                for part_index in range(longest):
                    processed_part = processed_line[part_index]
                    if processed_part is None:
                        continue
                    
                    only_none = False
                    if type(processed_part) is list:
                        processed_part = ' '.join(processed_part)
                    
                    processed_line[part_index] = processed_part
                
                if only_none:
                    continue
                
                processed_lines.append(processed_line)
                processed_line = [None for _ in range(longest)]
                continue
            
            # Add the current line to the actual processed line.
            for part_index in range(longest):
                splitted_part = splitted_line[part_index]
                if splitted_part is None:
                    continue
                
                processed_part = processed_line[part_index]
                if processed_part is None:
                    processed_line[part_index] = splitted_part
                elif type(processed_part) is list:
                    processed_part.append(splitted_part)
                else:
                    processed_line[part_index] = [processed_part, splitted_part]
        
        # Remove empty lines
        for index in reversed(range(len(processed_lines))):
            processed_line = processed_lines[index]
            if processed_line is None:
                continue
            
            for part in processed_line:
                if part:
                    break
            else:
                del processed_lines[index]
        
        # Remove None-s from end
        while True:
            if not processed_lines:
                return None
            
            if processed_lines[-1] is None:
                del processed_lines[-1]
                continue
            
            break
        
        # Remove None-s from start
        while True:
            if not processed_lines:
                return None
            
            if processed_lines[0] is None:
                del processed_lines[0]
                continue
            
            break
        
        # Remove Dupe None-s
        last_is_none = True
        for index in reversed(range(len(processed_lines))):
            processed_line = processed_lines[index]
            if processed_line is None:
                if last_is_none:
                    del processed_lines[index]
                else:
                    last_is_none = True
            else:
                last_is_none = False
        
        # Create size
        size = (longest, len(processed_lines))
        # Create array
        array = []
        for processed_line in processed_lines:
            array.extend(processed_line)
        
        self = object.__new__(cls)
        self._array = array
        self._size = size
        
        return self
    
    @property
    def size(self):
        """
        Returns the size of the table.
        
        Returns
        -------
        size : `tuple` (`int`, `int`)
            The x - y size of the table.
        """
        return self._size
    
    def __iter__(self):
        """
        Iterates over the table's lines.
        
        This method is a generator.
        
        Yields
        ------
        line : `list` of (`str`, `None`)
        """
        x, y = self._size
        array = self._array
        index = 0
        for _ in range(y):
            line = []
            for _ in range(x):
                element = array[index]
                index += 1
                
                line.append(element)
            
            yield line
    
    def __repr__(self):
        """Returns the table's representation."""
        result = [
            '<',
            self.__class__.__name__,
            ', size=',
            repr(self._size),
            ', table=[',
                ]
        
        for line in self:
            result.append(repr(line))
            result.append(', ')
    
        del result[-1]
        
        result.append(']>')
        return ''.join(result)
    
    def graved(self, path):
        """
        Returns a graved version of the table.
        
        Parameters
        ----------
        path : ``QualPath``
            The path of the parent docstring.
        
        Returns
        -------
        graved : ``GravedTable``
        """
        return GravedTable(self, path)

def detect_description(lines, index, limit):
    """
    Detects a description section.
    
    Parameters
    ----------
    lines : `list` of `str`
        The lines of the section.
    index : `int`
        The starting index of the description.
    limit : `int`
        The last element's index.
    
    Returns
    -------
    index : `int`
        The index where the description is over. If there was no description, returns the initial index.
    """
    while True:
        if index == limit:
            return index
        
        line = lines[index]
        
        if not line:
            return index
        
        if line[0] in (' ', '\t', '-'):
            return index
        
        if (TABLE_ANY_RP.fullmatch(line) is not None):
            return index
        
        index +=1
        continue


class TextDescription:
    """
    Represents a text part of a docstring.
    
    Attributes
    ----------
    _content : `str`
        The content of the description.
    """
    __slots__ = ('_content', )
    def __new__(cls, lines, start, end):
        """
        Creates a new description from the given lines's start:end range.
        
        Parameters
        ----------
        lines : `list` of `str`
            Lines to create the description from.
        start : `int`
            The first line's index, where the description starts.
        end : `int`
            The 1 after the last line's index of the description.
        """
        content = ' '.join(lines[start:end])
        
        self = object.__new__(cls)
        self._content = content
        return self
    
    def __repr__(self):
        """Returns the description's representation."""
        return f'<{self.__class__.__name__} content={self._content!r}>'
    
    def graved(self, path):
        """
        Returns a graved version of the description.
        
        Parameters
        ----------
        path : ``QualPath``
            The path of the parent docstring.
        
        Returns
        -------
        graved : ``GravedDescription``
        """
        return GravedDescription(self, path)

def detect_indent(lines, index, limit):
    """
    Detects an indented part.
    
    Parameters
    ----------
    lines : `list` of `str`
        The lines of the section.
    index : `int`
        The starting index of the indent.
    limit : `int`
        The last element's index.
    
    Returns
    -------
    index : `int`
        The index where the indent is over. If there was no indent, returns the initial index.
    """
    source_index = index
    found_indented = False
    while True:
        if index == limit:
            break
        
        line = lines[index]
        if not line:
            index +=1
            continue
        
        if line[0] in (' ', '\t'):
            found_indented = True
            index +=1
            continue
        
        break
    
    if found_indented:
        return index
    
    return source_index


def build_indent(lines, start, end):
    """
    Builds a section out from an indented part.
    
    Parameters
    ----------
    lines : `list` of `str`
        Lines to create the section from.
    start : `int`
        The first line's index, where the section starts.
    end : `int`
        The 1 after the last line's index of the section.
    
    Returns
    -------
    section : `None` or `list` of `Any`
        Instead of returning an empty section, returns `None`.
    """
    lines = lines[start:end]
    lines = remove_indents(lines)
    if lines is None:
        return None
    
    return parse_section(lines)


def detect_code_block(lines, index, limit):
    """
    Detects a code block and returns it's last line +1's index, or the source index if not found.
    
    Parameters
    ----------
    lines : `list` of `str`
        The lines of the section.
    index : `int`
        The starting index of the code block.
    limit : `int`
        The last element's index.
    
    Returns
    -------
    index : `int`
        The index where the code block is over. If there was no code block, returns the initial index.
    """
    source_index = index
    found_starter = False
    while True:
        if index == limit:
            if found_starter:
                return index
            else:
                return source_index
        
        line = lines[index]
        if not line:
            index += 1
            continue
        
        if line.endswith('```') if found_starter else line.startswith('```'):
            index += 1
            
            if found_starter:
                return index
            
            found_starter = True
            continue
        
        if found_starter:
            index += 1
            continue
        
        return source_index

class TextCodeBlock:
    """
    Represents a code-block part in a docstring.
    
    Attributes
    ----------
    _language : `None` or `str`
        The code block's language's name if any. Always lower case.
    _lines : `list` of `str`
        The internal lines of the code block.
    """
    __slots__ = ('_language', '_lines',)
    def __new__(cls, lines, start, end):
        """
        Creates a new code block from the given lines's start:end range.
        
        Parameters
        ----------
        lines : `list` of `str`
            Lines to create the code block from.
        start : `int`
            The first line's index, where the code block starts.
        end : `int`
            The 1 after the last line's index of the code block.

        
        Returns
        -------
        self : `None` or ``TextCodeBlock``
            Returns `None` if would create an empty code block.
        """
        line = lines[start]
        if len(line) == 3:
            language = None
            first_line = None
        else:
            first_line = line[3:]
            break_index = first_line.find(' ')
            if break_index == -1:
                language = first_line
                first_line = None
            else:
                language = first_line[:break_index]
                if language:
                    language = language.lower()
                else:
                    language = None
                
                first_line = first_line[break_index:].lstrip()
                if not first_line:
                    first_line = None
        
        if first_line is None:
            start += 1
        
        line = lines[end-1]
        if len(line) == 3:
            last_line = None
        else:
            last_line = line[:-3]
            last_line = last_line.rstrip(RIGHT_STRIP_IGNORE)
            if not last_line:
                last_line = None
        
        if last_line is None:
            end -= 1
        
        if start >= end:
            return None
        
        lines = lines[start:end]
        if (first_line is not None):
            lines[0] = first_line
        
        if (last_line is not None):
            lines[-1] = last_line
        
        # Remove empty lines from the end
        while True:
            if not lines:
                return None
            
            if lines[-1]:
                break
            
            del lines[-1]
            continue
        
        # Remove empty lines from the start.
        while True:
            if not lines:
                return None
            
            if lines[0]:
                break
            
            del lines[0]
            continue
        
        self = object.__new__(cls)
        self._lines = lines
        self._language = language
        return self
    
    def __repr__(self):
        """Returns the code blocks's representation."""
        result = ['<', self.__class__.__name__]
        
        language = self._language
        if language is None:
            result.append(' language=')
            result.append(repr(language))
            
            add_comma = True
        else:
            add_comma = False
        
        if add_comma:
            result.append(', ')
        
        result.append(' lines=')
        result.append(repr(self._lines))
        
        result.append('>')
        
        return ''.join(result)
    
    def graved(self, path):
        """
        Returns a graved version of the code block.
        
        Parameters
        ----------
        path : ``QualPath``
            The path of the parent docstring.
        
        Returns
        -------
        graved : ``GravedCodeBlock``
        """
        return GravedCodeBlock(self, path)

def detect_void(lines, index, limit):
    """
    Detects empty lines.
    
    Parameters
    ----------
    lines : `list` of `str`
        The lines of the section.
    index : `int`
        The starting index of the void..
    limit : `int`
        The last element's index.
    
    Returns
    -------
    index : `int`
        The index where the void is over. If there was no void, returns the initial index.
    """
    while True:
        if index == limit:
            return index
        
        line = lines[index]
        if line:
            return index
        
        index +=1
        continue


def build_void(lines, start, end):
    """
    Builds a void from the given parameters.
    
    Parameters
    ----------
    lines : `list` of `str`
        The lines of the section.
    start : `int`
        The starting index of the void.
    end : `int`
        The void's last line's index +1.
    
    Returns
    -------
    void : `None`
    """
    return None


def parse_section(lines):
    """
    Parses the given section and returns the built objects.
    
    Parameters
    ----------
    lines : `list` of `str`
        Lines to build the section from.
    
    Returns
    -------
    built : `None` or `list` of `Any`
        The built up elements of the sections. If would return an empty list, returns `None` instead.
    """
    index = 0
    limit = len(lines)
    
    builts = []
    
    while True:
        if index == limit:
            break
        
        for detector, builder in (
                (detect_void, build_void),
                (detect_indent, build_indent),
                (detect_table, TextTable),
                (detect_listing, TextListing),
                (detect_code_block, TextCodeBlock),
                (detect_block_quote, TextBlockQuote),
                (detect_description, TextDescription),
                    ):
            
            detected_end = detector(lines, index, limit)
            if detected_end == index:
                continue
            
            built = builder(lines, index, detected_end)
            index = detected_end
            if built is None:
                break
            
            builts.append(built)
            break
        else:
            sys.stderr.write(f'Could not detect anything, skipping. Line :{lines[index]}\n')
            index += 1
            continue
    
    if (not builts):
        return None
    
    # If a description is indented alone by 2 behind an other one, connect them.
    index = 0
    limit = len(builts)-1
    while True:
        if index >= limit:
            break
        
        built = builts[index]
        index +=1
        
        if type(built) is not TextDescription:
            continue
        
        next_built = builts[index]
        if type(next_built) is not list:
            continue
        
        next_built_internal = next_built[0]
        if type(next_built_internal) is not list:
            continue
        
        if len(next_built_internal) != 1:
            continue
        
        maybe_description = next_built_internal[0]
        if type(maybe_description) is not TextDescription:
            continue
        
        built._content = f'{built._content} {maybe_description._content}'
        del next_built[0]
        index +=1
        limit -=1
        continue
    
    return builts


def detect_listing(lines, index, limit):
    """
    Detects the lines of a listing.
    
    Parameters
    ----------
    lines : `list` of `str`
        The lines of the section.
    index : `int`
        The starting index of the listing.
    limit : `int`
        The last element's index.
    
    Returns
    -------
    index : `int`
        The index where the listing is over. If there was no listing, returns the initial index.
    """
    get_back_index = index
    dash_count = 0
    while True:
        if index == limit:
            return index
        
        line = lines[index]
        
        if line.startswith('-'):
            dash_count +=1
            index +=1
            continue
        
        if dash_count == 0:
            # No dash is detected
            return get_back_index
        
        if not line:
            index +=1
            continue
        
        if line[0] in (' ', '\t'):
            index +=1
            continue
        
        if dash_count == 1:
            return get_back_index
        
        break
    
    return index


class TextListingElement:
    """
    Represents an element of a listing.
    
    Attributes
    ----------
    _content : `None`, `list` of `Any`
        The content stored inside of a listing element. If would be set as empty `list`, is set as `None` instead.
    _head : `None` or `str`
        The head content of the listing. If would be set as empty string, is set as `None` instead.
    """
    __slots__ = ('_content', '_head')
    def __new__(cls, lines):
        """
        Creates a new listing element from the given lines.
        
        Parameters
        ----------
        lines : `list` of `str`
            The lines to create the element from.
        
        Returns
        -------
        self : `None` or ``TextListingElement``
            Instead of returning an empty listing element, returns `None` instead.
        """
        head_line = lines.pop(0)
        
        parsed = LISTING_HEAD_LINE_RP.fullmatch(head_line)
        if parsed is None:
            head = None
        else:
            head = parsed.group(1)
            if not head:
                head = None
        
        # GOTO
        content = None
        while True:
            if not lines:
                break
            
            continue_head = bool(lines[0])
            
            lines = remove_indents(lines)
            if lines is None:
                break
            
            section = parse_section(lines)
            if section is None:
                continue
            
            if not continue_head:
                content = section
                break
            
            maybe_description = section[0]
            if type(maybe_description) is not TextDescription:
                content = section
                break
            
            del section[0]
            
            extra_head_content = maybe_description._content
            if head is None:
                head = extra_head_content
            else:
                head = f'{head} {extra_head_content}'
            
            if not section:
                break
            
            content = section
            break
        
        if None is head is content:
            self = None
        else:
            self = object.__new__(cls)
            self._head = head
            self._content = content
        
        return self
    
    def __repr__(self):
        """Returns the listing element's representation."""
        result = [
            '<',
            self.__class__.__name__
                ]
        
        head = self._head
        if head is None:
            add_comma = False
        else:
            result.append(' head=')
            result.append(repr(head))
            add_comma = True
        
        content = self._content
        if (content is not None):
            if add_comma:
                result.append(',')
            
            result.append(' content=')
            result.append(repr(content))
        
        result.append('>')
        
        return ''.join(result)
    
    def graved(self, path):
        """
        Returns a graved version of the listing element.
        
        Parameters
        ----------
        path : ``QualPath``
            The path of the parent docstring.
        
        Returns
        -------
        graved : ``GravedListingElement``
        """
        return GravedListingElement(self, path)


class TextListing:
    """
    Represents a listing inside of a docstring.
    
    Attributes
    ----------
    _elements : `list` of ``TextListingElement``
        The elements of the listing.
    """
    __slots__ = ('_elements')
    def __new__(cls, lines, start, end):
        """
        Creates a new listing instance.
        
        Parameters
        ----------
        lines : `list` of `str`
            Lines to build the listing from.
        start : `int`
            The first lines index of the listing.
        end : `int`
            The last lines +1 index.
        
        Returns
        -------
        self : `None` or ``TextListing``
            If would create an empty listing, returns `None` instead.
        """
        sections = []
        
        # Separate each part
        section_parts = []
        while True:
            if start == end:
                if section_parts:
                    sections.append(section_parts)
                break
            
            line = lines[start]
            start +=1
            
            if line.startswith('-'):
                # If the line starts with `-`, then save the currently created part if any.
                if section_parts:
                    sections.append(section_parts)
                    section_parts = []
            
            section_parts.append(line)
            continue
        
        # Remove empty lines from the end and remove empty sections if any.
        index = len(sections)
        
        while True:
            if index == 0:
                break
            
            index -= 1
            section_parts = sections[index]
            
            while True:
                if not section_parts:
                    del sections[index]
                    break
                
                if not section_parts[-1]:
                    del section_parts[-1]
                    continue
                
                break
        
        # Nice, now lets make the elements.
        elements = []
        for section_parts in sections:
            element = TextListingElement(section_parts)
            if element is None:
                continue
            
            elements.append(element)
            continue
        
        if not elements:
            return None
        
        self = object.__new__(cls)
        self._elements = elements
        return self
    
    def __repr__(self):
        """Returns the listing's representation."""
        return f'<{self.__class__.__name__} elements={self._elements!r}>'
    
    def graved(self, path):
        """
        Returns a graved version of the listing.
        
        Parameters
        ----------
        path : ``QualPath``
            The path of the parent docstring.
        
        Returns
        -------
        graved : ``GravedListing``
        """
        return GravedListing(self, path)


def detect_block_quote(lines, index, limit):
    """
    Detects the lines of a block quote.
    
    Parameters
    ----------
    lines : `list` of `str`
        The lines of the section.
    index : `int`
        The starting index of the block quote.
    limit : `int`
        The last element's index.
    
    Returns
    -------
    index : `int`
        The index where the block quote is over. If there was no block quote, returns the initial index.
    """
    while True:
        if index == limit:
            return index
        
        line = lines[index]
        if not line:
            return index
        
        if line[0] not in '>':
            return index
        
        index += 1
        continue

def remove_block_quote_indents(lines):
    """
    Removes the dedent from the given lines returning a new list with the lines without them.
    
    Parameters
    ----------
    lines : `list` of `str`
        Input lines.
    
    Returns
    -------
    lines : `None` or (`list` of `str`)
    """
    if not lines:
        return None
    
    for index in range(len(lines)):
        lines[index] = lines[index][1:].lstrip()
    
    while True:
        if lines[-1]:
            break
        
        del lines[-1]
        if not lines:
            return None
    
    while True:
        if lines[0]:
            break
        
        del lines[0]
        continue
    
    return lines


class TextBlockQuote:
    """
    Represents a block quote of a docstring.
    
    Attributes
    ----------
    _descriptions : `list` of  ``TextDescription``
        The description parts inside of the block quote.
    """
    __slots__ = ('_descriptions', )
    def __new__(cls, lines, start, end):
        """
        Creates a new block quote from the given lines's start:end range.
        
        Parameters
        ----------
        lines : `list` of `str`
            Lines to create the block quote from.
        start : `int`
            The first line's index, where the block quote starts.
        end : `int`
            The 1 after the last line's index of the block quote..
        """
        lines = lines[start:end]
        lines = remove_block_quote_indents(lines)
        if lines is None:
            return None
        
        index = 0
        limit = len(lines)
        
        builts = []
        
        while True:
            if index == limit:
                break
        
        
            for detector, builder in (
                    (detect_void, build_void),
                    (detect_description, TextDescription),
                        ):
                
                detected_end = detector(lines, index, limit)
                if detected_end == index:
                    continue
                
                built = builder(lines, index, detected_end)
                index = detected_end
                if built is None:
                    break
                
                builts.append(built)
                break
            else:
                sys.stderr.write(f'Could not detect anything, skipping. Line :{lines[index]}\n')
                index += 1
                continue
        
        if not builts:
            return None
        
        self = object.__new__(cls)
        self._descriptions = builts
        return self
    
    def __repr__(self):
        """Returns the description's representation."""
        return f'<{self.__class__.__name__} descriptions={self._descriptions!r}>'
    
    def graved(self, path):
        """
        Returns a graved version of the block quote.
        
        Parameters
        ----------
        path : ``QualPath``
            The path of the parent docstring.
        
        Returns
        -------
        graved : ``GravedBlockQuote``
        """
        return GravedBlockQuote(self, path)


def parse_docstring(text, path):
    """
    Parses the given docstring.
    
    Parameters
    ---------
    text : `None` or `str`
        The docstring itself.
    path : ``QUalPath``
        The path of the docstring's object.
    
    Returns
    -------
    built_sections : `None` or `list` of (`tuple` ((`None` or `str`), `Any`))
    """
    if text is None:
        return None
    
    lines = text.splitlines()
    
    remove_indents(lines)
    if lines is None:
        DocWarning(path, 'Empty docstring not defined as `None`.')
        return None
    
    if lines[0][0] in (' ', '\t'):
        DocWarning(path, 'Docstring starts with space or tab character. Might contain `\\n` or mixed indents.')
        return None
    
    sections = parse_sections(lines)
    if (sections[0][0] is not None):
        DocWarning(path, 'No general description provided.')
    
    built_sections = []
    
    for section_name, section_lines in sections:
        built = parse_section(section_lines)
        if built is None:
            DocWarning(path, f'Empty subsection part: {section_name}')
            continue
        
        built_sections.append((section_name, built))
    
    if (not built_sections):
        return None
    
    result = []
    
    for section_name, sub_section in built_sections:
        sub_section = build_graves_on_subsection(sub_section, path)
        if sub_section is None:
            DocWarning(path, f'Empty subsection part: {section_name}')
            continue
        
        result.append((section_name, sub_section))
    
    if not result:
        return None
    
    return result


def convert_extra_attribute_section_name(name):
    return '___'+name.lower().replace(' ', '_')

def get_attribute_docs_from(sections):
    """
    Gets the attribute doc parts from the given sections.
    
    Parameters
    ----------
    sections : `list` of `tuple` ((`None` or `str`), (`list` of `Any`))
        The sections of the docstring.
    
    Returns
    -------
    result : `dict` of (`str`, ``DocString``) items
    """
    result = {}
    
    for name, section in sections:
        if name is None:
            continue
        
        if ATTRIBUTE_SECTION_NAME_RP.fullmatch(name) is None:
            continue
        
        index = 0
        limit = len(section)
        while True:
            if index == limit:
                break
            
            part = section[index]
            if (type(part) is not GravedDescription):
                break
            
            starter = part.content[0]
            if (type(starter) is not str):
                break
            
            parsed = ATTRIBUTE_NAME_RP.match(starter)
            if parsed is None:
                break
            
            attr_name, attr_separator = parsed.groups()
            attr_head = part.content.copy()
            starter_continuous = starter[parsed.end():]
            if starter_continuous:
                attr_head[0] = starter_continuous
            else:
                if len(attr_head) == 1:
                    attr_head = []
                else:
                    del attr_head[0]
            
            index += 1
            if index == limit:
                attr_body = None
            else:
                part = section[index]
                if type(part) is list:
                    attr_body = part
                    index += 1
                else:
                    attr_body = None
            
            result[attr_name] = DocString._create_attribute_docstring_part(attr_name, attr_separator, attr_head, attr_body)
            continue
        
        extra = []
        while True:
            if index == limit:
                break
            
            part = section[index]
            index += 1
            extra.append(part)
            continue
        
        if extra:
            name = convert_extra_attribute_section_name(name)
            result[name] = DocString._create_attribute_docstring_extra(extra)
    
    return result

class DocString:
    """
    Represents a docstring.
    
    Attributes
    ----------
    _attribute_sections : `None` or `dict` of (`str`, ``DocString``) items
        A dict containing the attribute sections of the docstring. Set only when an attribute's docstring is requested.
    sections : `list` of `tuple` ((`None` or `str`), (`list` of `Any`))
        The sections of the docstring.
    """
    __slots__ = ('_attribute_sections', 'sections')
    def __new__(cls, docstring, path):
        """
        Creates a new docstring.
        
        docstring : `None` or ``Docstring``
        """
        if docstring is None:
            return None
        
        sections = parse_docstring(docstring, path)
        if sections is None:
            return None
        
        self = object.__new__(cls)
        self._attribute_sections = None
        self.sections = sections
        return self
    
    def __repr__(self):
        """Returns the docstring's representation."""
        return f'<{self.__class__.__name__} sections={self.sections!r}>'
    
    @classmethod
    def _create_attribute_docstring_part(cls, attr_name, attr_separator, attr_head, attr_body):
        """
        Creates an attribute docstring.
        
        Parameters
        ----------
        attr_name : `str`
            The name of the respective attribute.
        attr_separator : `str`
            Separator used between the attribute's name and it's type description.
        attr_head : `None` or `str`
            The attribute description following the attribute's name.
        attr_body : `None` or `list` of `Any`
            Extra content after the attribute's head.
        
        Returns
        -------
        self : ``DocString``
        """
        head = GravedAttributeDescription(attr_name, attr_separator, attr_head)
        
        section = [head]
        
        if (attr_body is not None):
            section.extend(attr_body)
        
        self = object.__new__(cls)
        self._attribute_sections = None
        self.sections = [(None, section)]
        return self
    
    @classmethod
    def _create_attribute_docstring_extra(cls, attr_body):
        """
        Creates an attribute docstring.
        
        Parameters
        ----------
        attr_body : `None` or `list` of `Any`
            The extra content.
        
        Returns
        -------
        self : ``DocString``
        """
        self = object.__new__(cls)
        self._attribute_sections = None
        self.sections = [(None, attr_body)]
        return self
    
    def attribute_docstring_for(self, name):
        """
        Gets the docstring for the given attribute.
        
        Parameters
        ----------
        name : `str`
            The name of the attribute.
        
        Returns
        -------
        docstring : `None` or ``Docstring``
        """
        attribute_sections = self._attribute_sections
        if attribute_sections is None:
            self._attribute_sections = attribute_sections = get_attribute_docs_from(self.sections)
        
        return attribute_sections.get(name, None)
    
    def extra_attribute_docstring_for(self, name):
        """
        Gets the extra docstring for the given attribute section name.
        
        Parameters
        ----------
        name : `str`
            The name of the attribute section.
        
        Returns
        -------
        docstring : `None` or ``Docstring``
        """
        attribute_sections = self._attribute_sections
        if attribute_sections is None:
            self._attribute_sections = attribute_sections = get_attribute_docs_from(self.sections)
        
        name = convert_extra_attribute_section_name(name)
        return attribute_sections.get(name, None)



# -*- coding: utf-8 -*-
__all__ = ('ClaimedExecutor', 'Executor', 'ExecutorThread', 'SyncQueue', 'SyncWait', )

import sys
from threading import Lock as SyncLock, Event as SyncEvent, Thread, current_thread
from collections import deque

from .export import include
from .futures import Future, _ignore_frame, render_exc_to_list, CancelledError

EventThread = include('EventThread')

_ignore_frame(__spec__.origin, 'wait', 'raise exception', )
_ignore_frame(__spec__.origin, 'result_no_wait', 'raise exception', )
_ignore_frame(__spec__.origin, 'run', 'result=func()', )

class SyncWait:
    """
    Blocking waiter, what's result or exception can be set.
    
    Attributes
    ----------
    _exception : `None` or `BaseException` instance
        The waiter's exception if applicable. defaults to `None`.
    _result : `Any`
        The waiter's result if applicable. Defaults to `None`.
    _waiter : `threading.SyncEvent`
        Threading event, what is set, when the waiter's result is set.
    """
    __slots__ = ('_exception', '_result', '_waiter',)
    def __init__(self):
        """
        Creates a new ``SyncWait`` instance.
        """
        self._exception = None
        self._result = None
        self._waiter = SyncEvent()
    
    def set_exception(self, exception):
        """
        Sets an exception as the waiter's result.
        
        Parameters
        ----------
        exception : `BaseException`
            The exception to set as the future's exception.
        
        Raises
        ------
        TypeError
            If `StopIteration` is given as `exception`.
        """
        if isinstance(exception, type):
            exception = exception()
        
        if type(exception) is StopIteration:
             raise TypeError(f'{exception} cannot be raised to a {self.__class__.__name__}: {self!r}')
        
        if self._exception is None:
            self._exception = exception
            self._waiter.set()
        else:
            if type(exception) is CancelledError and type(self._exception) is not CancelledError:
                self._exception = exception
    
    def set_result(self, result):
        """
        Sets the waiter's result.
        
        Parameters
        ----------
        result : `Any`
            The object to set as result.
        """
        self._result = result
        self._waiter.set()
    
    def wait(self):
        """
        Waits till the waiter's result is set, then returns it.
        
        Returns
        -------
        result : `Any`
        
        Raises
        ------
        TypeError
            The set exception is not a  `BaseException` instance.
        BaseException
            The set exception.
        """
        self._waiter.wait()
        exception = self._exception
        if exception is None:
            return self._result
        raise exception
    
    def cancel(self):
        """
        Cancels the waiter by setting it's exception as ``CancelledError`` if not yet set.
        """
        exception = self._exception
        if exception is None:
            self._waiter.set()
        else:
            if type(exception) is CancelledError:
                return
        
        self._exception = CancelledError()
        

class SyncQueue:
    """
    Synchronous queue for retrieving set elements outside of an event loop.
    
    Attributes
    ----------
    _cancelled : `bool`
        Whether the queue is cancelled.
    _lock : `threading.SyncLock`
        Threading lock to grant exclusive access to the queue.
    _results : `deque`
        A deque of the results set to the queue.
    _waiter : `None` or ``SyncWait``
        Result waiter of the queue.
    """
    __slots__ = ('_cancelled', '_lock', '_results', '_waiter',)
    
    def __init__(self, iterable=None, max_length=None, cancelled=False):
        """
        Creates a new ``SyncQueue`` instance with the given parameters.
        
        Parameters
        ----------
        iterable : `iterable` of `Any`
            Iterable to set the queue's results initially from.
        max_length : `int`
            Maximal length of the queue. If the queue would pass it's maximal length, it's oldest results are popped.
        cancelled : `bool`
            Whether the queue should be initially cancelled.
        """
        self._results = deque(maxlen=max_length) if iterable is None else deque(iterable, maxlen=max_length)
        self._waiter = None
        self._cancelled = cancelled
        self._lock = SyncLock()
    
    def set_result(self, result):
        """
        Sets an element to the sync queue, what can be retrieved by ``.result`` or  by ``.result_no_wait``.
        
        Parameters
        ----------
        result : `Any`
            Sets a result to the sync queue.
        """
        with self._lock:
            # Should we raise InvalidStateError?
            waiter = self._waiter
            if waiter is None:
                self._results.append(result)
            else:
                waiter.set_result(result)
                self._waiter = None
    
    def cancel(self):
        """
        Cancels the sync-queue and it's waiter as well.
        """
        if self._cancelled:
            return
        
        with self._lock:
            # Should we raise InvalidStateError?
            self._cancelled = True
            
            waiter = self._waiter
            if waiter is not None:
                waiter.cancel()
                self._waiter = None
    
    def result(self):
        """
        Gets the next result of the sync queue. If the queue is empty, blocks till a new result is set with
        ``.set_result``.
        
        Returns
        -------
        result : `Any`
        """
        with self._lock:
            results = self._results
            if results:
                return results.popleft()
            
            waiter = self._waiter
            if waiter is None:
                if self._cancelled:
                    raise CancelledError()
                
                waiter = SyncWait()
                self._waiter = waiter
        
        return waiter.wait()
    
    def result_no_wait(self):
        """
        Gets the next result of the sync queue instantly.
        
        Returns
        -------
        result : `Any`
        
        Raises
        ------
        CancelledError
            If the queue is empty and cancelled.
        IndexError
            If the queue is empty, but not cancelled.
        """
        with self._lock:
            results = self._results
            if results:
                return results.popleft()
            
            if self._cancelled:
                exception = CancelledError()
            else:
                raise IndexError('The queue is empty.')
            
            raise exception
    
    wait = result
    
    def __repr__(self):
        """Returns the sync queue's representation."""
        with self._lock:
            result = [
                self.__class__.__name__,
                '([',
            ]
            
            results = self._results
            
            limit = len(results)
            if limit:
                index = 0
                while True:
                    element = results[index]
                    index += 1
                    
                    results.append(repr(element))
                    if index == limit:
                        break
                    
                    result.append(', ')
                    continue
            
            result.append(']')
            
            max_length = results.maxlen
            if (max_length is not None):
                result.append(', max_length=')
                result.append(repr(max_length))
            
            if self._cancelled:
                result.append(', cancelled')
            
            result.append(')')
            return ''.join(result)
    
    __str__ = __repr__
    
    # deque operations
    
    @property
    def max_length(self):
        """
        The maximal length of the queue if applicable.
        
        Returns
        -------
        max_length : `None` or `int`
        """
        return self._results.maxlen
    
    def clear(self):
        """
        Clears the queue.
        """
        self._results.clear()
    
    def copy(self):
        """
        Copies the queue.
        
        Returns
        -------
        new : ``SyncQueue``
        """
        with self._lock:
            new = object.__init__(type(self))
            
            new._results = self._results.copy()
            new._waiter = None
            new._cancelled = self._cancelled
            new._lock = SyncLock()
            
        return new
    
    def reverse(self):
        """
        Reverses the queue.
        """
        with self._lock:
            self._results.reverse()
    
    def __len__(self):
        """Returns the queue's length."""
        with self._lock:
            return len(self._results)
    
    def __bool__(self):
        """Returns `True` if the queue has any elements."""
        with self._lock:
            if self._results:
                result = True
            else:
                result = False
        
        return result

EXECUTOR_THREAD_CREATED = 0
EXECUTOR_THREAD_RUNNING = 1
EXECUTOR_THREAD_STOPPED = 2

class ExecutorThread(Thread):
    """
    A `threading.Thread` subclass for executing blocking code as an executor.
    
    Attributes
    ----------
    state : `int`
        Whether the executor thread is currently running.
        
        Can be on of the following:
        +---------------------------+-------+
        | Respective name           | Value |
        +===========================+=======+
        | EXECUTOR_THREAD_CREATED   | 0     |
        +---------------------------+-------+
        | EXECUTOR_THREAD_RUNNING   | 1     |
        +---------------------------+-------+
        | EXECUTOR_THREAD_STOPPED   | 2     |
        +---------------------------+-------+
    
    queue : ``SyncQueue`` of ``ExecutionPair``
        Synchronous queue of functions to execute and of their waiter future.
    """
    __slots__ = ('state', 'queue',)
    def __init__(self):
        """
        Creates and start a ``ExecutorThread``.
        """
        self.state = EXECUTOR_THREAD_CREATED
        self.queue = SyncQueue()
        Thread.__init__(self, daemon=True)
        self.start()
        
    def run(self):
        """
        The main runner of an ``ExecutorThread``.
        """
        if self.state != EXECUTOR_THREAD_CREATED:
            return
            
        self.state = EXECUTOR_THREAD_RUNNING
        
        queue = self.queue
        
        while self.state == EXECUTOR_THREAD_RUNNING:
            try:
                try:
                    pair = queue.wait()
                except CancelledError:
                    return
                
                future = pair.future
                if future.done():
                    future = None
                    continue
                
                func = pair.func
                try:
                    result = func()
                except BaseException as err:
                    if isinstance(err, StopIteration):
                        exception = RuntimeError(f'{err.__class__.__name__} cannot be raised to a Future')
                        exception.__cause__ = err
                        err = exception
                        exception = None
                    
                    future._loop.call_soon_thread_safe(future.__class__.set_exception_if_pending, future, err)
                else:
                    future._loop.call_soon_thread_safe(future.__class__.set_result_if_pending, future, result)
                    result = None
                
                future = None
                func = None
            
            except BaseException as err:
                extracted = [
                    self.__class__.__name__,
                    ' exception occurred\n',
                    repr(self),
                    '\n',
                        ]
                render_exc_to_list(err, extend=extracted)
                sys.stderr.write(''.join(extracted))
    
    def execute(self, func, future=None):
        """
        Executes the given `func` on the executor thread.
        
        Parameters
        ----------
        func : `callable`
            The function to execute.
        future : `None` or ``Future``, Optional
            A future, what's result is set, when the `func` finishes it's execution.
            
            If not given a new future is created. Defaults to `None`
        
        Returns
        -------
        future : ``Future``
            Result waiter future.
        
        Raises
        ------
        RuntimeError
            `future` parameter is not given an the method was not called from an event thread.
        """
        if future is None:
            loop = current_thread()
            if not isinstance(loop, EventThread):
                raise RuntimeError(f'`future` parameter is not given and `{self!r}.execute` was not called '
                    f'from an{EventThread.__name__} either.')
            
            future = Future(loop)
        
        self.queue.set_result(ExecutionPair(func, future,),)
        return future
    
    def cancel(self):
        """
        Cancels all the pending elements on the executor thread's queue, then cancels the queue itself as well.
        """
        self.state = EXECUTOR_THREAD_STOPPED
        queue = self.queue
        while queue:
            future = queue.result_no_wait().future
            future._loop.call_soon_thread_safe(future.cancel)
        
        queue.cancel()
    
    def release(self):
        """
        Cancels the executor thread's queue.
        """
        self.queue.cancel()
    
    def _shrink_queue(self):
        """
        Removes all the done elements from the executor's queue.
        """
        results = self.queue._results
        
        index = 0
        limit = len(results)
        while index < limit:
            future = results[index]
            if future.done():
                del results[index]
                limit -= 1
            else:
                index += 1

class ExecutionPair:
    """
    Stores a function to execute and it's future, to what the function's result or exception is set when done.
    
    Attributes
    ----------
    func : `callable`
        The function to execute.
    future : ``Future``
        The result waiter future to what ``.func``'s result or exception is set.
    """
    __slots__ = ('func', 'future', )
    
    def __init__(self,func,future):
        """
        Creates a new ``ExecutionPair`` instance.
        
        Parameters
        ----------
        func : `callable`
            The function to execute.
        future : ``Future``
            The result waiter future to what ``.func``'s result or exception is set.
        """
        self.func = func
        self.future = future
    
    def __repr__(self):
        """Returns the execution pair's representation."""
        return f'{self.__class__.__name__}(func={self.func!r}, future={self.future!r})'

class _claim_ended_cb:
    """
    Future callback set to result waiter futures when calling ``ClaimedExecutor.release``.
    
    Gives back the executor thread, when it's queue becomes empty.
    
    Parameters
    ----------
    parent : ``Executor`` instance
        The parent executor
    executor : ``ExecutorThread``
        The executor thread, what is given to the parent executor when all of it's tasks are done.
    """
    __slots__ = ('executor', 'parent',)
    def __init__(self, parent, executor):
        """
        Creates a new ``_claim_ended_cb`` instance with the given parameters.
        
        Parameters
        ----------
        parent : ``Executor`` instance
            The parent executor
        executor : ``ExecutorThread``
            The executor thread, what is given to the parent executor when all of it's tasks are done.
        """
        self.parent = parent
        self.executor = executor
    
    def __call__(self, future):
        """Gives back the executor thread, when it's queue becomes empty."""
        loop = future._loop
        executor = self.executor
        executor._shrink_queue()
        if not executor.queue:
            parent = self.parent
            loop.call_soon_thread_safe(parent.__class__._claim_ended, parent, executor)
    
    def __eq__(self,other):
        """Returns whether the two claim ended callbacks are the same."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self.executor is other.executor
    
class ClaimedExecutor:
    """
    Wrapper for claimed executor threads.
    
    Attributes
    ----------
    executor : ``ExecutorThread``
        The claimed executor thread.
    parent : ``Executor``
        The parent executor from where the ``.executor`` is claimed.
    """
    __slots__ = ('executor', 'parent',)
    def __init__(self, parent, executor):
        """
        Creates a new claimed executor instance which will wrap the given executor thread of it's executor.
        
        Parameters
        ----------
        parent : ``Executor``
            The parent executor from where the ``.executor`` is claimed.
        executor : ``ExecutorThread``
            The claimed executor thread.
        """
        self.parent = parent
        self.executor = executor
        
    def execute(self, func):
        """
        Runs the given function in an executor thread. When the function is done, it's result or exception is set to
        the returned future.
        
        Parameters
        ----------
        func : ``Callable``
            The function to run inside of an executor thread.
        
        Returns
        -------
        future : ``Future``
            The future, to what the returned value or the raised exception of `func` is set.
        
        Raises
        ------
        RuntimeError
            The claimed executor is closed.
        """
        executor = self.executor
        if executor is None:
            raise RuntimeError(f'Executing on an already closed {self.__class__.__name__}.')
        
        future = self.parent.create_future()
        
        executor.queue.set_result(ExecutionPair(func, future,),)
        
        return future

    def release(self):
        """
        Releases the claimed executor. Gives back it's executor thread to the parent executor, when all of it's
        tasks finish.
        """
        executor = self.executor
        if executor is None:
            return
        
        executor._shrink_queue()
        
        results = executor.queue._results
        if results:
            callback = _claim_ended_cb(self.parent, executor)
            for element in results:
                future = element.future
                if future.done():
                    continue
                
                future.add_done_callback(callback)
        else:
            self.parent._claim_ended(executor)
        
        self.executor = None
    
    __del__ = release
    
    def __enter__(self):
        """Entering a claimed executor returns itself."""
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        """Releases the claimed executor."""
        self.release()
    
class _execution_ended_cb:
    """
    Future callback set to result waiter futures when calling ``Executor.run_in_executor``.
    
    Parameters
    ----------
    parent : ``Executor`` instance
        The parent executor
    executor : ``ExecutorThread``
        The executor thread, what is given to the parent executor when it's current task is done.
    """
    __slots__ = ('executor', 'parent',)
    
    def __init__(self, parent, executor):
        """
        Creates a new ``_execution_ended_cb`` instance with the given parameters.
        
        Parameters
        ----------
        parent : ``Executor`` instance
            The parent executor
        executor : ``ExecutorThread``
            The executor thread, what is given to the parent executor when it's current task is done.
        """
        self.parent = parent
        self.executor = executor
        
    def __call__(self, future):
        """calls the parent executor's ``._execution_ended`` method, giving the executor thread back to it."""
        parent = self.parent
        future._loop.call_soon_thread_safe(parent.__class__._execution_ended, parent, self.executor)
    
    def __eq__(self, other):
        """Returns whether the two execution ended callbacks are the same."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self.executor is other.executor

class Executor:
    """
    Executor enabling running blocking sync code from async.
    
    Attributes
    ----------
    claimed_executors : `set` of ``ExecutorThread``
        Claimed executors, which are given back to the executor on release.
    free_executors : `deque`
        The free (or not used) executors of the executor.
    keep_executor_count : `int`
        The minimal amount of executors to keep alive (or not close).
    running_executors : `set` of ``ExecutorThread``
        The executors under use.
    """
    __slots__ = ('claimed_executors', 'free_executors', 'keep_executor_count', 'running_executors',)
    
    def __init__(self, keep_executor_count=1):
        """
        Initializes the executor.
        
        Parameters
        ----------
        keep_executor_count : `int`, Optional
            The minimal amount of executors to keep alive (or not close).
        """
        self.free_executors = deque()
        self.running_executors = set()
        self.claimed_executors = set()
        self.keep_executor_count = keep_executor_count
    
    def __repr__(self):
        """Returns the executor's representation."""
        return (f'<{self.__class__.__name__} free={self.free_executor_count}, used={self.used_executor_count}, keep='
            f'{self.keep_executor_count}>')
    
    @property
    def used_executor_count(self):
        """
        Returns how much executor threads of the executors are currently under use.
        
        Returns
        -------
        used_executor_count : `int`
        """
        return len(self.running_executors)+len(self.claimed_executors)
    
    @property
    def free_executor_count(self):
        """
        Returns how much free executors
        """
        return len(self.free_executors)
    
    def cancel_executors(self):
        """
        Cancels the executor threads of the executor.
        
        Raises ``CancelledError`` to the not yet started tasks.
        """
        self.keep_executor_count = 0
        executors = self.free_executors
        while executors:
            executor = executors.pop()
            executor.release()
        
        executors = self.running_executors
        while executors:
            executor = executors.pop()
            executor.cancel()
        
        executors = self.claimed_executors
        while executors:
            executor = executors.pop()
            executor.cancel()
    
    def release_executors(self):
        """
        Releases the executor threads of the executor.
        
        Raises no exception to the already started tasks,
        """
        self.keep_executor_count = 0
        executors = self.free_executors
        while executors:
            executor = executors.pop()
            executor.release()
        
        executors = self.running_executors
        while executors:
            executor = executors.pop()
            executor.release()
        
        executors = self.claimed_executors
        while executors:
            executor = executors.pop()
            executor.release()
    
    __del__ = release_executors
    
    def create_future(self):
        """
        Creates a future bound to the respective event loop.
        
        Returns
        -------
        future : ``Future``
            The created future.
        
        Raises
        ------
        RuntimeError
            If the method was not called from an ``SyncEventLoop``.
        """
        local_thread = current_thread()
        if not isinstance(local_thread, EventThread):
            raise RuntimeError(f'{self!r}.create_future was not called from {EventThread.__name__}, but from '
                f'{local_thread.__class__.__name__}')
        
        return Future(local_thread)
    
    def run_in_executor(self, func):
        """
        Runs the given function in an executor thread. When the function is done, it's result or exception is set to
        the returned future.
        
        Parameters
        ----------
        func : ``Callable``
            The function to run inside of an executor thread.
        
        Returns
        -------
        future : ``Future``
            The future, to what the returned value or the raised exception of `func` is set.
            
            If the executor is cancelled, then these futures are cancelled as well.
        """
        future = self.create_future()
        executor = self._get_free_executor()
        self.running_executors.add(executor)
        future.add_done_callback(_execution_ended_cb(self, executor))
        executor.queue.set_result(ExecutionPair(func, future,),)
        return future

    def _execution_ended(self, executor):
        """
        Called when an executed function finishes ensured by ``.run_in_executor``. Used to give back the
        ``ExecutorThread`` to the executor and maybe cancel it.
        
        Parameters
        ----------
        executor : ``ExecutorThread``
            Executor with finished execution.
        """
        try:
            self.running_executors.remove(executor)
        except KeyError:
            return
        
        self._sync_keep(executor)
    
    def claim_executor(self):
        """
        Claims an executor thread from the executor. The thread is given back to the executor, when the returned
        wrapper is cancelled or released.
        
        Returns
        -------
        claimed_executor : ``ClaimedExecutor``
        """
        executor = self._get_free_executor()
        self.claimed_executors.add(executor)
        return ClaimedExecutor(self, executor)
    
    def _claim_ended(self, executor):
        """
        Called when a claimed executor is released and all of it's tasks are done. Used to give back the
        ``ExecutorThread`` to the executor and maybe cancel it.
        
        Parameters
        ----------
        executor : ``ExecutorThread``
            The given back executor.
        """
        try:
            self.claimed_executors.remove(executor)
        except KeyError:
            return
        
        self._sync_keep(executor)
    
    def _get_free_executor(self):
        """
        Gets a free executor thread from the executor. If there are no free executor threads, starts a new one.
        
        Returns
        -------
        executor : ``ExecutorThread``
        """
        executors = self.free_executors
        if executors:
            executor = executors.pop()
        else:
            executor = ExecutorThread()
        
        return executor

    def _sync_keep(self, executor):
        """
        Called when an executor thread is given back to the executor.
        
        The method decides whether the executor should be kept or released.
        
        Parameters
        ----------
        executor : ``ExecutorThread``
            The given back executor.
        """
        executors = self.free_executors
        if len(executors) < self.keep_executor_count:
            executors.append(executor)
            return
        
        executor.release()



# -*- coding: utf-8 -*-
from ipaddress import ip_address
from urllib.parse import SplitResult, parse_qsl as parse_query_string_list, urljoin as url_join, \
    urlsplit as url_split, urlunsplit as url_unsplit
from math import isinf, isnan

from .utils import multidict, cached_property
from .quote import quote, unquote

NoneType = type(None)

DEFAULT_PORTS = {
    'http': 80,
    'https': 443,
    'ws': 80,
    'wss': 443,
}

class URL:
    """
    Represents an URL (Uniform Resource Locator)
    
    Attributes
    ----------
    _cache : `dict` of (`str`, `str`) items
        Internal cache used by ``cached_property``-s.
    _value : `urllib.parse.SplitResult`
        Internal value of the ``URL``.
    """
    # Don't derive from str follow pathlib.Path design probably URL will not suffer from pathlib problems: it's
    # intended for libraries, not to be passed into standard library functions like os.open etc.

    # URL grammar (RFC 3986)
    # pct-encoded = "%" HEXDIG HEXDIG
    # reserved = gen-delims / sub-delims
    # gen-delims = ":" / "/" / "?" / "#" / "[" / "]" / "@"
    # sub-delims = "!" / "$" / "&" / "'" / "(" / ")"
    #             / "*" / "+" / "," / ";" / "="
    # unreserved = ALPHA / DIGIT / "-" / "." / "_" / "~"
    # URI = scheme ":" hier-part [ "?" query ] [ "#" fragment ]
    # hier-part = "//" authority path-abempty
    #             / path-absolute
    #             / path-rootless
    #             / path-empty
    # scheme = ALPHA *( ALPHA / DIGIT / "+" / "-" / "." )
    # authority = [ userinfo "@" ] host [ ":" port ]
    # userinfo = *( unreserved / pct-encoded / sub-delims / ":" )
    # host = IP-literal / IPv4address / reg-name
    # IP-literal = "[" ( IPv6address / IPvFuture  ) "]"
    # IPvFuture = "v" 1*HEXDIG "." 1*( unreserved / sub-delims / ":" )
    # IPv6address = 6( h16 ":" ) ls32
    #             /                       "::" 5( h16 ":" ) ls32
    #             / [               h16 ] "::" 4( h16 ":" ) ls32
    #             / [ *1( h16 ":" ) h16 ] "::" 3( h16 ":" ) ls32
    #             / [ *2( h16 ":" ) h16 ] "::" 2( h16 ":" ) ls32
    #             / [ *3( h16 ":" ) h16 ] "::"    h16 ":"   ls32
    #             / [ *4( h16 ":" ) h16 ] "::"              ls32
    #             / [ *5( h16 ":" ) h16 ] "::"              h16
    #             / [ *6( h16 ":" ) h16 ] "::"
    # ls32 = ( h16 ":" h16 ) / IPv4address
    #             ; least-significant 32 bits of address
    # h16 = 1*4HEXDIG
    #             ; 16 bits of address represented in hexadecimal
    # IPv4address = dec-octet "." dec-octet "." dec-octet "." dec-octet
    # dec-octet = DIGIT                 ; 0-9
    #             / %x31-39 DIGIT         ; 10-99
    #             / "1" 2DIGIT            ; 100-199
    #             / "2" %x30-34 DIGIT     ; 200-249
    #             / "25" %x30-35          ; 250-255
    # reg-name = *( unreserved / pct-encoded / sub-delims )
    # port = *DIGIT
    # path = path-abempty    ; begins with "/" or is empty
    #               / path-absolute   ; begins with "/" but not "//"
    #               / path-noscheme   ; begins with a non-colon segment
    #               / path-rootless   ; begins with a segment
    #               / path-empty      ; zero characters
    # path-abempty = *( "/" segment )
    # path-absolute = "/" [ segment-nz *( "/" segment ) ]
    # path-noscheme = segment-nz-nc *( "/" segment )
    # path-rootless = segment-nz *( "/" segment )
    # path-empty = 0<pchar>
    # segment = *pchar
    # segment-nz = 1*pchar
    # segment-nz-nc = 1*( unreserved / pct-encoded / sub-delims / "@" )
    #               ; non-zero-length segment without any colon ":"
    # pchar = unreserved / pct-encoded / sub-delims / ":" / "@"
    # query = *( pchar / "/" / "?" )
    # fragment = *( pchar / "/" / "?" )
    # URI-reference = URI / relative-ref
    # relative-ref = relative-part [ "?" query ] [ "#" fragment ]
    # relative-part = "//" authority path-abempty
    #               / path-absolute
    #               / path-noscheme
    #               / path-empty
    # absolute-URI = scheme ":" hier-part [ "?" query ]
    
    __slots__ = ('_cache', '_value', )

    def __new__(cls, value='', encoded=False):
        """
        Creates a new ``URL`` instance from the given `value`
        
        Parameters
        ----------
        value : ``URL``, `str`, `urllib.parse.SplitResult`, Optional
            The value to create ``URL`` from. Defaults to empty string.
        encoded : `bool`, Optional
            Whether the given `value` is already encoded. Defaults to `False`.

        Raises
        -------
        ValueError
            - If `value` is given as `urllib.parse.SplitResult` instance, but `encoded` was given as `False`.
            - If `value` is not `encoded` and the URL is absolute, but `host` is not given.
        TypeError
            If `value` was not given neither as ``URL``, `str` nor `urllib.parse.SplitResult` instance.
        """
        if isinstance(value, cls):
            return value
        
        if isinstance(value, str):
            value = url_split(value)
        elif isinstance(value, SplitResult):
            if not encoded:
                raise ValueError(f'Cannot apply decoding to `{SplitResult.__name__}`.')
        else:
            raise TypeError(f'`value` should have be given as `{cls.__name__}`, `str` or `{SplitResult.__name__}` '
                f'instance, got {value.__class__.__name__}.')
        
        if not encoded:
            if value.netloc:
                netloc = value.hostname
                if netloc is None:
                    raise ValueError('Invalid URL: host is required for absolute urls.')
                
                try:
                    netloc.encode('ascii')
                except UnicodeEncodeError:
                    netloc = netloc.encode('idna').decode('ascii')
                else:
                    try:
                        ip = ip_address(netloc)
                    except:
                        pass
                    else:
                        if ip.version == 6:
                            netloc = f'[{netloc}]'
                
                value_port = value.port
                if value_port:
                    netloc = f'{netloc}:{value_port}'
                
                value_username = value.username
                if value_username:
                    user = quote(value_username)
                    value_password = value.password
                    if value_password:
                        user = f'{user}:{quote(value_password)}'
                    
                    netloc = f'{user}@{netloc}'
            else:
                netloc = ''
            
            value = SplitResult(value.scheme, netloc,
                quote(value.path, safe='@:', protected='/'),
                quote(value.query, safe='=+&?/:@', protected='=+&', query_string=True),
                quote(value.fragment, safe='?/:@'),
                    )
        
        self = object.__new__(cls)
        self._value = value
        self._cache = {}
        return self
    
    def __str__(self):
        """Returns str(self)."""
        value = self._value
        if not value.path and self.is_absolute() and (value.query or value.fragment):
            value = value._replace(path='/')
        
        return url_unsplit(value)
    
    def __repr__(self):
        """Returns repr(self)."""
        return f'{self.__class__.__name__}({str(self)!r})'

    def __hash__(self):
        """Returns hash(self)."""
        try:
            hash_value = self._cache['hash']
        except KeyError:
            hash_value = self._cache['hash'] = hash(self._value)
        return hash_value
    
    def __gt__(self, other):
        """Returns (self > other)."""
        if type(self) is type(other):
            pass
        elif isinstance(other, str):
            other = type(self)(other)
        else:
            return NotImplemented
        
        return self._value > other._value
        
    def __ge__(self, other):
        """Returns (self >= other)."""
        if type(self) is type(other):
            pass
        elif isinstance(other, str):
            other = type(self)(other)
        else:
            return NotImplemented
        
        return self._value >= other._value
        
    def __eq__(self, other):
        """Returns (self == other)."""
        if type(self) is type(other):
            pass
        elif isinstance(other, str):
            other = type(self)(other)
        else:
            return NotImplemented
        
        return self._value == other._value
        
    def __ne__(self, other):
        """Returns (self != other)."""
        if type(self) is type(other):
            pass
        elif isinstance(other, str):
            other = type(self)(other)
        else:
            return NotImplemented
        
        return self._value != other._value
    
    def __le__(self, other):
        """Returns (self <= other)."""
        if type(self) is type(other):
            pass
        elif isinstance(other, str):
            other = type(self)(other)
        else:
            return NotImplemented
        
        return self._value <= other._value
        
    def __lt__(self, other):
        """Returns (self < other)."""
        if type(self) is type(other):
            pass
        elif isinstance(other, str):
            other = type(self)(other)
        else:
            return NotImplemented
        
        return self._value < other._value
    
    def __truediv__(self, name):
        """Returns self / other"""
        name = quote(name, safe=':@', protected='/')
        if name.startswith('/'):
            raise ValueError('Appending path starting from slash is forbidden')
        path = self._value.path
        if path == '/':
            new_path = f'/{name}' 
        elif not path and not self.is_absolute():
            new_path = name
        else:
            parts = path.rstrip('/').split('/')
            parts.append(name)
            new_path = '/'.join(parts)
        return URL(self._value._replace(path=new_path, query='', fragment=''), encoded=True)
    
    def is_absolute(self):
        """
        Returns whether the URL is absolute (having scheme or starting with //).
        
        Returns
        -------
        is_absolute : `bool`
        """
        return (self.raw_host is not None)
    
    def is_default_port(self):
        """
        Returns whether the URL's port is default, like: 'http://python.org' or 'http://python.org:80'.
        
        Returns
        -------
        is_default_port : `bool`
        """
        if self.port is None:
            return False
        
        default = DEFAULT_PORTS.get(self.scheme, None)
        if default is None:
            return False
        
        if self.port == default:
            return True
        
        return False
    
    def origin(self):
        """
        Returns an URL with scheme, host and port parts only, user, password, path, query and fragment are removed.
        
        Returns
        -------
        new_url : ``URL``
        
        Raises
        ------
        ValueError
            - If the URL is not absolute.
            - If the URL has no scheme.
        """
        if not self.is_absolute():
            raise ValueError('URL should be absolute.')
        
        value = self._value
        if not value.scheme:
            raise ValueError('URL should have scheme.')
        
        netloc = self._make_netloc(None, None, value.hostname, value.port)
        new_value = value._replace(netloc=netloc, path='', query='', fragment='')
        return URL(new_value, encoded=True)
    
    def relative(self):
        """
        Returns a relative part of the URL, scheme, user, password, host and port are removed.
        
        Returns
        -------
        new_url : ``URL``
        
        Raises
        ------
        ValueError
            If URL is not absolute.
        """
        if not self.is_absolute():
            raise ValueError('URL should be absolute.')
        
        value = self._value._replace(scheme='', netloc='')
        return URL(value, encoded=True)
    
    @property
    def scheme(self):
        """
        Returns the scheme for absolute URL-s. Returns empty string for relative URL-s and for URL-s starting with //.
        
        Returns
        -------
        scheme : `str`
        """
        return self._value.scheme
    
    @property
    def raw_user(self):
        """
        Returns the encoded user part of the URL. `None` is returned if the user part is missing.
        
        Returns
        -------
        raw_user : `None` or `str`
        """
        return self._value.username
    
    @cached_property
    def user(self):
        """
        Returns the decoded user part of the URL. `None` is returned if the user part is missing.
        
        Returns
        -------
        user : `None` or `str`
        """
        return unquote(self.raw_user)
    
    @property
    def raw_password(self):
        """
        Returns the encoded password part of the URL. Returns `None` if password is missing.
        
        Returns
        -------
        raw_password : `None` or `str`
        """
        return self._value.password
    
    @cached_property
    def password(self):
        """
        Returns the decoded password part of the URL. Returns `None` if password is missing.
        
        Returns
        -------
        password : `None` or `str`
        """
        return unquote(self.raw_password)
    
    @property
    def raw_host(self):
        """
        Returns the encoded host part of the URL. Returns `None` if the host part is missing or if the URL is relative.
        
        Returns
        -------
        raw_host : `None` or `str`
        """
        return self._value.hostname
    
    @cached_property
    def host(self):
        """
        Returns the decoded host part of the URL. Returns `None` if the host part is missing or if the URL is relative.
        
        Returns
        -------
        host : `None` or `str`
        """
        raw_host = self.raw_host
        if raw_host is None:
            host = None
        else:
            host = raw_host.encode('ascii').decode('idna')
        
        return host
    
    @property
    def raw_subdomain(self):
        """
        Returns the encoded subdomain part of the URL if it has. Als returns `None` if the host part is missing or if
        the URL is relative.
        
        Returns
        -------
        raw_subdomain : `None` or `str`
        """
        host_name = self._value.hostname
        if host_name is None:
            subdomain = None
        else:
            back_index = host_name.rfind('.')
            if back_index == -1:
                subdomain = None
            else:
                back_index = host_name.rfind('.', back_index)
                if back_index == -1:
                    subdomain = None
                else:
                    subdomain = host_name[:back_index]
        
        return subdomain
        
    @cached_property
    def subdomain(self):
        """
        Returns the decoded subdomain of the URL if it has. Also returns `None` if the host part is missing or if the
        Url is relative.
        
        Returns
        -------
        subdomain : `None` or `str`
        """
        raw_subdomain = self.raw_subdomain
        if raw_subdomain is None:
            subdomain = None
        else:
            subdomain = raw_subdomain.encode('ascii').decode('idna')
        
        return subdomain
    
    @property
    def port(self):
        """
        Returns the port part of URL. Returns `None` if the URL is relative, if the URL not contains port part and if
        the port can's be detected from the URL's scheme.
        
        Returns
        -------
        port : `None` or `int`
        """
        value = self._value
        port = value.port
        if port is None:
            port = DEFAULT_PORTS.get(value.scheme, None)
        
        return port
    
    @property
    def raw_path(self):
        """
        Returns the encoded path part of the URL. Returns `'/'` for absolute URL-s without any path parts.
        
        Returns
        -------
        raw_path : `str`
        """
        raw_path = self._value.path
        if (not raw_path) and self.is_absolute():
            raw_path = '/'
        
        return raw_path
    
    @cached_property
    def path(self):
        """
        Returns the decoded path part of the URL. Returns `'/'` for absolute URL-s without any path parts.
        
        Returns
        -------
        path : `str`
        """
        return unquote(self.raw_path)
    
    @cached_property
    def query(self):
        """
        A multidict representing parsed query parameters in decoded representation.
        
        An empty multidict is returned if the url has no query parts.
        
        Returns
        -------
        query : `multidict` of (`str`, `Any`) items
        """
        return multidict(parse_query_string_list(self.query_string, keep_blank_values=True))
    
    @property
    def raw_query_string(self):
        """
        Returns the encoded query string part of URL. Returns an empty string if the query part is missing.
        
        Returns
        -------
        raw_query_string : `str`
        """
        return self._value.query
    
    @cached_property
    def query_string(self):
        """
        Returns the decoded query string part of URL. Returns an empty string if the query part is missing.
        
        Returns
        -------
        query_string : `str`
        """
        return unquote(self.raw_query_string, query_string=True)
    
    @property
    def raw_fragment(self):
        """
        Returns the encoded fragment part of the URL. Returns an empty string if the fragment part is missing.
        
        Returns
        -------
        raw_fragment : `str`
        """
        return self._value.fragment

    @cached_property
    def fragment(self):
        """
        Returns the decoded fragment part of the URL. Returns an empty string if the fragment part is missing.
        
        Returns
        -------
        raw_fragment : `str`
        """
        return unquote(self.raw_fragment)
    
    @cached_property
    def raw_parts(self):
        """
        Returns a `tuple` containing encoded path parts.
        
        If the URL is absolute, or if path is missing, it's first element always will be `'/'`.

        Returns
        -------
        raw_parts : `tuple` of `str`
        """
        path = self._value.path
        
        parts = []
        if self.is_absolute():
            parts.append('/')
            if path:
                parts.extend(path[1:].split('/'))
        else:
            if path.startswith('/'):
                parts.append('/')
                parts.extend(path[1:].split('/'))
            else:
                parts.extend(path.split('/'))
        
        return tuple(parts)
    
    @cached_property
    def parts(self):
        """
        Returns a `tuple` containing decoded path parts.
        
        If the URL is absolute, or if path is missing, it's first element always will be `'/'`.

        Returns
        -------
        raw_parts : `tuple` of `str`
        """
        return tuple(unquote(part) for part in self.raw_parts)
    
    @cached_property
    def parent(self):
        """
        Returns a new URL with last part of path removed without query and fragment.
        
        Returns
        -------
        parent : ``URL``
        """
        path = self.raw_path
        if path == '/':
            if self.raw_fragment or self.raw_query_string:
                parent = URL(self._value._replace(query='', fragment=''), encoded=True)
            else:
                parent = self
        else:
            parts = path.split('/')
            value = self._value._replace(path='/'.join(parts[:-1]), query='', fragment='')
            parent = URL(value, encoded=True)
        
        return parent
    
    @cached_property
    def raw_name(self):
        """
        Returns the last part of ``.raw_parts``. If there are no parts, returns an empty string.
        
        Returns
        -------
        raw_name : `str`
        """
        parts = self.raw_parts
        if self.is_absolute() and (len(parts) < 2):
            raw_name = ''
        else:
            raw_name = parts[-1]
        
        return raw_name
    
    @cached_property
    def name(self):
        """
        Returns the last part of ``.parts``. If there are no parts, returns an empty string.
        
        Returns
        -------
        raw_name : `str`
        """
        return unquote(self.raw_name)
    
    @classmethod
    def _make_netloc(cls, user, password, host, port):
        """
        Makes netloc from the given parameters
        
        Parameters
        ----------
        user : `None` or `str`
            User part of an URL.
        password : `None` or `str`
            Password part of an URL.
        host : `None` or `str`
            The host part of an URL.
        port : `None` or `int`
            The ort part of an URL:
        
        Returns
        -------
        netloc : `str`
        
        Raises
        ------
        ValueError
            If password is not empty, `user` parameter should not be empty either.
        """
        if (password is not None) and password:
            if (user is None) or (not user):
                raise ValueError('Non-empty password requires non-empty user.')
            
            user = f'{user}:{password}'
        
        if host is None:
            netloc = ''
        else:
            netloc = host
        
        if (port is not None) and port:
            netloc = f'{netloc}:{port!s}'
        
        if (user is not None) and user:
            netloc = f'{user}@{netloc}'
        
        return netloc
    
    def with_scheme(self, scheme):
        """
        Returns a new URL with `scheme` replaced.
        
        Parameters
        ----------
        scheme : `str`
            Scheme part for the new url.

        Returns
        -------
        new_url : ``URL``
        
        Raises
        ------
        TypeError
            If `scheme` was not given as `str` instance.
        ValueError
            If the source URL is relative; Scheme replacement is not allowed for relative URL-s.
        
        Notes
        -----
        The returned URL's `query` and `fragment` will be same as the source one's.
        """
        if not isinstance(scheme, str):
            raise TypeError(f'`scheme` can be given as `str` instance, got {scheme.__class__.__name__}.')
        
        if not self.is_absolute():
            raise ValueError('`scheme` replacement is not allowed for relative URL-s.')
        
        return URL(self._value._replace(scheme=scheme.lower()), encoded=True)
    
    def with_user(self, user):
        """
        Returns a new URL with `user` replaced.
        
        Give `user` as `None` to clear `user` and `password` from the source URL.
        
        Parameters
        ----------
        user : `None` or `str`
            User part for the new url.

        Returns
        -------
        new_url : ``URL``
        
        Raises
        ------
        TypeError
            If `user` was not given neither as `None` nor `str` instance.
        ValueError
            If the source URL is relative; User replacement is not allowed for relative URL-s.
        
        Notes
        -----
        The returned URL's `query` and `fragment` will be same as the source one's.
        """
        value = self._value
        if user is None:
            password = None
        elif isinstance(user, str):
            user = quote(user)
            password = value.password
        else:
            raise TypeError(f'`user` can be given as `None` or `str` instance, got  {user.__class__.__name__}.')
        
        if not self.is_absolute():
            raise ValueError('`user` replacement is not allowed for relative URL-s.')
        
        return URL(value._replace(netloc=self._make_netloc(user, password, value.hostname, value.port)), encoded=True)
    
    def with_password(self, password):
        """
        Returns a new URL with `password` replaced.
        
        Give `password` as `None` to clear it from the source URL.
        
        Parameters
        ----------
        password : `None` or `str`
            Password part for the new url.

        Returns
        -------
        new_url : ``URL``
        
        Raises
        ------
        TypeError
            If `password` was not given neither as `None` nor `str` instance.
        ValueError
            If the source URL is relative; Password replacement is not allowed for relative URL-s.
        
        Notes
        -----
        The returned URL's `query` and `fragment` will be same as the source one's.
        """
        if password is None:
            pass
        elif isinstance(password, str):
            password = quote(password)
        else:
            raise TypeError(f'`password` can be given as `None` or `str` instance, got {password.__class__.__name__}.')
        
        if not self.is_absolute():
            raise ValueError('Password replacement is not allowed for relative URL-s.')
        
        value = self._value
        return URL( value._replace(netloc=self._make_netloc(value.username, password, value.hostname, value.port)),
            encoded=True)
    
    def with_host(self, host):
        """
        Returns a new URL with `host` replaced.
        
        Parameters
        ----------
        host : `str`
            Host part for the new url.

        Returns
        -------
        new_url : ``URL``
        
        Raises
        ------
        TypeError
            If `host` was not given as `str` instance.
        ValueError
            - If the source URL is relative; Host replacement is not allowed for relative URL-s.
            - If `host` was given as empty string; Removing `host` is not allowed.
        
        Notes
        -----
        The returned URL's `query` and `fragment` will be same as the source one's.
        """
        if not isinstance(host, str):
            raise TypeError(f'`host` can be given as `str` instance, got {host.__class__.__name__}.')
        
        if not self.is_absolute():
            raise ValueError('`host` replacement is not allowed for relative URL-s.')
        
        if not host:
            raise ValueError('`host` was given as empty string, but removing host is not allowed.')
        
        try:
            ip = ip_address(host)
        except:
            host = host.encode('idna').decode('ascii')
        else:
            if ip.version == 6:
                host = f'[{host}]'
        
        value = self._value
        return URL(value._replace(netloc=self._make_netloc(value.username, value.password, host, value.port)),
            encoded=True)
    
    def with_port(self, port):
        """
        Returns a new URL with `port` replaced. Give `port` as None` to clear it to default.
        
        Parameters
        ----------
        port : `int` or `None`
            Port part of the new url.

        Returns
        -------
        new_url : ``URL``
        
        Raises
        ------
        TypeError
            If `port` was not given neither as `None` nor `int` instance.
        ValueError
            If the source URL is relative; Port replacement is not allowed for relative URL-s.
        
        Notes
        -----
        The returned URL's `query` and `fragment` will be same as the source one's.
        """
        if (port is not None) and (not isinstance(port, int)):
            raise TypeError(f'`port` can be given as `None` or `int` instance, got {port.__class__.__name__}.')
        
        if not self.is_absolute():
            raise ValueError('`port` replacement is not allowed for relative URL-s.')
        
        value = self._value
        
        return URL(value._replace(netloc=self._make_netloc(value.username, value.password, value.hostname, port)),
            encoded=True)
    
    def with_query(self, query):
        """
        Returns a new url with query part replaced. By giving `None` you can clear the actual query.
        
        Parameters
        ----------
        query : `None`, `str`, (`dict`, `list`, `set`) of \
                (`str`, (`str`, `int`, `bool`, `NoneType`, `float`, (`list`, `set`, `tuple`) of repeat value)) items
            The query to use.
        
        Returns
        -------
        new_url : ``URL``
        
        Raises
        ------
        TypeError
            - If `query` was given as an invalid type.
            - If `query` was given as `set` or `list`, but `1` of it's elements cannot be unpacked correctly.
            - If a query key was not given as `str` instance.
            - If a query value was not given as any of the expected types.
        ValueError
            - If a query value was given as `float`, but as `inf`.
            - If a query value was given as `float`, but as `nan`.
        
        Notes
        -----
        The returned URL's `fragment` will be same as the source one's.
        """
        if query is None:
            query = ''
        elif isinstance(query, dict):
            query = build_query_from_dict(query)
        elif isinstance(query, (list, set)):
            query = build_query_from_list(query)
        elif isinstance(query, str):
            query = quote(query, safe='/?:@', protected='=&+', query_string=True)
        elif isinstance(query, (bytes, bytearray, memoryview)):
            raise TypeError(f'Query type cannot be `bytes`, `bytearray` and `memoryview`, got '
                f'{query.__class__.__name__}')
        else:
            raise TypeError(f'Invalid query type: {query.__class__.__name__}.')
        
        value = self._value
        
        path = value.path
        if not path:
            path = '/'
        
        return URL(value._replace(path=path, query=query), encoded=True)
    
    def with_fragment(self, fragment):
        """
        Returns a new URL with `fragment` replaced. Give `fragment` as None` to clear it.
        
        Parameters
        ----------
        fragment : `str` or `None`
            Fragment part of the new url.

        Returns
        -------
        new_url : ``URL``
        
        Raises
        ------
        TypeError
            If `fragment` was not given neither as `None` nor `str` instance.
        
        Notes
        -----
        The returned URL's `query` will be same as the source one's.
        """
        if fragment is None:
            fragment = ''
        elif not isinstance(fragment, str):
            raise TypeError(f'`fragment` can be given as `None` or `str` instance, got {fragment.__class__.__name__}.')
        
        return URL(self._value._replace(fragment=quote(fragment, safe='?/:@')), encoded=True)
    
    def with_name(self, name):
        """
        Returns a new URL with `name` (last part of path) replaced.
        
        Parameters
        ----------
        name : `str`
            name part of the new url.

        Returns
        -------
        new_url : ``URL``
        
        Raises
        ------
        TypeError
            If `name` was not given as `str` instance.
        ValueError
            If `name` contains `'/'` character.
        
        Notes
        -----
        The returned URL's `fragment` and `query` will be REMOVED.
        """
        if not isinstance(name, str):
            raise TypeError(f'`name` can be given as `str` instance, got {name.__class__.__name__}.')
        
        if '/' in name:
            raise ValueError(f'`name` contains `\'/\'` character, got {name!r}.')
        
        name = quote(name, safe='@:', protected='/')
        
        parts = list(self.raw_parts)
        if self.is_absolute():
            if len(parts) == 1:
                parts.append(name)
            else:
                parts[-1] = name
            parts[0] = '' # replace leading '/'
        else:
            parts[-1] = name
            if parts[0] == '/':
                parts[0] = '' # replace leading '/'
        
        return URL(self._value._replace(path='/'.join(parts), query='', fragment=''), encoded=True)
    
    def join(self, other):
        """
        Joins two urls.
        
        Construct a full ('absolute') URL by combining a 'base URL' (self) with another one (other).
        
        Informally, this uses components of the base URL, in particular the addressing scheme, the network location and
        (part of) the path, to provide missing components in the relative URL.
        
        Parameters
        ----------
        other : ``URL``
            The other url to join to self.
        
        Raises
        ------
        TypeError
            If `other` was not given as ``URL`` instance.
        
        See Also
        -----
        `urllib.parse.url_join` for more information.
        """
        if not isinstance(other, URL):
            raise TypeError(f'`url` can be given as `{URL.__name__}` instance, got {other.__class__.__name__}.')
        
        return URL(url_join(str(self), str(other)), encoded=True)
    
    def human_repr(self):
        """
        Returns a hooman readable string representing the URL.
        
        Returns
        -------
        human_repr : `str`
        """
        return url_unsplit(SplitResult(self.scheme, self._make_netloc(self.user, self.password, self.host,
            self._value.port), self.path, self.query_string, self.fragment))
    
    def extend_query(self, params):
        """
        Returns a new URL with it's query parameters extended.
        
        Parameters
        ----------
        params : `None`, `str`, (`dict`, `list`, `set`) of \
                (`str`, (`str`, `int`, `bool`, `NoneType`, `float`, (`list`, `set`, `tuple`) of repeat value)) items
            The query parameters to extend the actual.
        
        Returns
        -------
        new_url : ``URL``
        
        Raises
        ------
        TypeError
            - If `query` was given as an invalid type.
            - If `query` was given as `set` or `list`, but `1` of it's elements cannot be unpacked correctly.
            - If a query key was not given as `str` instance.
            - If a query value was not given as any of the expected types.
        ValueError
            - If a query value was given as `float`, but as `inf`.
            - If a query value was given as `float`, but as `nan`.
        
        Notes
        -----
        The returned URL's `fragment` will be same as the source one's.
        """
        if (params is None) or (not params):
            new_url = self
        else:
            query = self.query.copy()
            if isinstance(params, str):
                params = parse_query_string_list(params, keep_blank_values=True)
            
            query.extend(params)
            
            new_url = self.with_query(query)
        
        return new_url


def build_query_from_dict(query):
    """
    Builds a query string and adds the parts of it to the given `build_to` list.
    
    Parameters
    ----------
    query : `dict` of (`str`, (`str`, `int`, `bool`, `NoneType`, `float`, (`list`, `tuple`, `set`) of repeat value)) items
        The query to serialize.
    
    Returns
    -------
    query_string : `str`
        The built query string.
    
    Raises
    ------
    TypeError
        - If `1` of `query`'s keys is not `str` instance.
        - If a query value is not any of the expected types.
    ValueError
        - If a query value was given as `float` and it is `inf`.
        - If a query value was given as `float` and it is `nan`.
    """
    build_to = []
    
    for key, value in query.items():
        query_key = quote(key, safe="/?:@", query_string=True)
        build_query_element_to(build_to, query_key, value)
    
    return ''.join(build_to)


def build_query_from_list(query):
    """
    Builds a query string and adds the parts of it to the given `build_to` list.
    
    Parameters
    ----------
    query : (`list` or `set`) of `tuple` \
            (`str`, (`str`, `int`, `bool`, `NoneType`, `float`, (`list`, `tuple`, `set`) of repeat value))
        The query to serialize.
    
    Returns
    -------
    query_string : `str`
        The built query string.
    
    Raises
    ------
    TypeError
        - If `1` of `query`'s elements cannot be unpacked to a `key` - `value` pair.
        - If `1` of `query`'s keys is not `str` instance.
        - If a query value is not any of the expected types.
    ValueError
        - If a query value was given as `float` and it is `inf`.
        - If a query value was given as `float` and it is `nan`.
    """
    build_to = []
    
    for key, value in query:
        query_key = quote(key, safe="/?:@", query_string=True)
        build_query_element_to(build_to, query_key, value)
    
    return ''.join(build_to)


def build_query_element_to(build_to, query_key, value):
    """
    Builds a query string element to the given `build_to` list.
    
    Parameters
    ----------
    build_to : `list`
        A list to build the query element to.
    query_key : `str`
        An already escaped query key.
    value : `str`, `int`, `bool`, `NoneType`, `float`, (`list`, `tuple`,  `set`) of repeat
        The query string value.
    
    Raises
    ------
    TypeError
        - If `value` was not given as any of the expected types.
    ValueError
        - If `value` is a `float`, but it is `inf`.
        - If `value` is a `float`, but it is `nan`.
    """
    if isinstance(value, str):
        query_value = value
    elif isinstance(value, int):
        query_value = str(value)
    elif isinstance(value, bool):
        query_value = 'true' if value else 'false'
    elif isinstance(value, NoneType):
        query_value = 'null'
    elif isinstance(value, float):
        if isinf(value):
             raise ValueError('`inf` is not a supported query string parameter value.')
        
        if isnan(value):
            raise ValueError('`nan` is not a supported query string parameter value.')
        
        query_value = str(value)
    
    elif isinstance(value, (list, tuple, set)):
        build_query_list_to(build_to, query_key, value)
        return
    
    else:
        raise TypeError(f'Unexpected value type received when serializing query string, got '
            f'{value.__class__.__name__}; {value!r}.')
    
    query_value = quote(query_value, safe="/?:@", query_string=True)
    
    if build_to:
        build_to.append('&')
    build_to.append(query_key)
    build_to.append('=')
    build_to.append(query_value)


def build_query_list_to(build_to, query_key, query_list):
    """
    Builds a query element from a `list` or `set`.
    
    Parameters
    ----------
    build_to : `list`
        A list to build the query element to.
    query_key : `str`
        An already escaped query key.
    query_list : (`list`, `tuple`, `set`) of (`str`, `int`, `bool`, `NoneType`, `float`, repeat)
        The query string value.
    
    Raises
    -----
    TypeError
        - If `query_list` contains a value with an unexpected type.
    ValueError
        - If `query_list` contains a value as a `float` what it is `inf`.
        - If `query_list` contains a value as a `float` what it is `nan`.
    """
    for value in query_list:
        build_query_element_to(build_to, query_key, value)
